import numpy as np
import pytest
from torch.utils.data import DataLoader

from continuum.datasets import InMemoryDataset
from continuum.scenarios import ClassIncremental


class InMemoryDatasetTest(InMemoryDataset):

    def __init__(self, *args, class_order=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._class_order = class_order

    @property
    def class_order(self):
        return self._class_order


@pytest.fixture
def numpy_data():
    x_train = []
    y_train = []
    for i in range(10):
        x_train.append(np.ones((5, 4, 4, 3), dtype=np.uint8) * i)
        y_train.append(np.ones(5) * i)
    x_train = np.concatenate(x_train)
    y_train = np.concatenate(y_train)

    x_test = np.copy(x_train)
    y_test = np.copy(y_train)

    return (x_train, y_train.astype(int)), (x_test, y_test.astype(int))


# yapf: disable

@pytest.mark.parametrize("classes,default_class_order,class_order", [
    ([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]], None, None),
    ([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]], list(range(10)), None),
    ([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]], None, list(range(10))),
    ([[1, 3, 5, 7, 9], [0, 2, 4, 6, 8]], [1, 3, 5, 7, 9, 0, 2, 4, 6, 8], None),
    ([[1, 3, 5, 7, 9], [0, 2, 4, 6, 8]], None, [1, 3, 5, 7, 9, 0, 2, 4, 6, 8]),
    ([[1, 3, 5, 7, 9], [0, 2, 4, 6, 8]], list(range(10)), [1, 3, 5, 7, 9, 0, 2, 4, 6, 8])
])
def test_increments(numpy_data, classes, default_class_order, class_order):
    train, test = numpy_data
    dummy = InMemoryDatasetTest(*train, class_order=default_class_order)
    scenario = ClassIncremental(dummy, 2, 5, class_order=class_order)

    gt_new_targets = [np.arange(5), np.arange(5) + 5]
    for task_id, taskset in enumerate(scenario):
        for _ in DataLoader(taskset):
            pass

        unique_classes = np.sort(np.unique(taskset._x))
        ans = (unique_classes == np.array(classes[task_id]))
        assert ans.all(), (task_id, unique_classes, np.array(classes[task_id]))

        original_targets = np.sort(np.unique(scenario.get_original_targets(taskset._y)))
        ans = (original_targets == np.array(classes[task_id]))
        assert ans.all(), (task_id, original_targets, np.array(classes[task_id]))

        new_targets = np.sort(np.unique(taskset._y))
        ans = (new_targets == gt_new_targets[task_id])
        assert ans.all(), (task_id, new_targets, gt_new_targets[task_id])



import numpy as np
import pytest

from continuum.datasets import InMemoryDataset
from continuum.scenarios import ClassIncremental, InstanceIncremental


@pytest.fixture
def numpy_data():
    nb_classes = 6
    nb_tasks = 3
    nb_data = 100

    x_train = []
    y_train = []
    t_train = []
    for i in range(nb_tasks):
        for j in range(nb_classes):
            for _ in range(nb_data):
                x_train.append(np.array(["hello world"]))
            y_train.append(np.ones(nb_data) * j)
            t_train.append(np.ones(nb_data) * i)

    x_train = np.concatenate(x_train)
    y_train = np.concatenate(y_train)
    t_train = np.concatenate(t_train)

    x_test = np.copy(x_train)
    y_test = np.copy(y_train)
    t_test = np.copy(t_train)

    return (x_train, y_train.astype(int), t_train), (x_test, y_test.astype(int), t_test)


def test_nlp_class_incremental(numpy_data):
    train, test = numpy_data

    x_train, y_train, t_train = train

    dummy = InMemoryDataset(x_train, y_train, data_type="text")

    scenario = ClassIncremental(dummy, increment=2)

    assert len(scenario) == scenario.nb_tasks == 3
    assert scenario.nb_classes == 6

    for _ in scenario:
        pass


def test_nlp_instance_incremental(numpy_data):
    train, test = numpy_data

    x_train, y_train, t_train = train

    dummy = InMemoryDataset(
        x_train, y_train, t=t_train, data_type="text"
    )

    scenario = InstanceIncremental(dummy)

    assert len(scenario) == scenario.nb_tasks == 3
    assert scenario.nb_classes == 6

    for _ in scenario:
        pass



import numpy as np
import pytest
from torchvision.transforms import ToTensor
from torch.utils.data import DataLoader
import torch

from continuum.datasets import InMemoryDataset
from continuum.scenarios import ClassIncremental, InstanceIncremental, ContinualScenario


@pytest.fixture
def dataset():
    x = np.random.randint(0, 255, (100, 4, 4, 3), dtype=np.uint8)
    y = np.random.randint(0, 3, (100,), dtype=np.int16)
    t = np.ones_like(y)

    t[:30] = 0
    t[30:60] = 1
    t[60:] = 2

    return InMemoryDataset(x, y, t)



@pytest.mark.parametrize("scenario,opt", [
    (ClassIncremental, {'increment': 1}),
    (InstanceIncremental, {}),
    (ContinualScenario, {})
])
def test_same_transforms(dataset, scenario, opt):
    trsfs = [
        ToTensor(),
        lambda tensor: tensor.fill_(0)
    ]
    s = scenario(dataset, transformations=trsfs, **opt)

    for taskset in s:
        loader = DataLoader(taskset)
        for x, _, _ in loader:
            assert torch.unique(x).numpy().tolist() == [0]


@pytest.mark.parametrize("scenario,opt", [
    (ClassIncremental, {'increment': 1}),
    (InstanceIncremental, {}),
    (ContinualScenario, {})
])
def test_diff_transforms(dataset, scenario, opt):
    trsfs = [
        [ToTensor(), lambda tensor1: tensor1.fill_(0)],
        [ToTensor(), lambda tensor2: tensor2.fill_(1)],
        [ToTensor(), lambda tensor3: tensor3.fill_(2)],
    ]
    s = scenario(dataset, transformations=trsfs, **opt)

    for taskid, taskset in enumerate(s):
        loader = DataLoader(taskset)
        for x, _, _ in loader:
            assert torch.unique(x).numpy().tolist() == [taskid]


@pytest.mark.parametrize("scenario,opt,error", [
    (ClassIncremental, {'increment': 1}, True),
    (InstanceIncremental, {}, True),
    (ContinualScenario, {}, True),
    (ClassIncremental, {'increment': 1}, False),
    (InstanceIncremental, {}, False),
    (ContinualScenario, {}, False)
])
def test_diff_transforms_slice(dataset, scenario, opt, error):
    trsfs = [
        [ToTensor(), lambda tensor1: tensor1.fill_(0)],
        [ToTensor(), lambda tensor2: tensor2.fill_(1)],
        [ToTensor(), lambda tensor3: tensor3.fill_(2)],
    ]
    s = scenario(dataset, transformations=trsfs, **opt)

    for taskid in range(len(s)):
        if not error:
            taskset = s[taskid]
            loader = DataLoader(taskset)
            for x, _, _ in loader:
                assert torch.unique(x).numpy().tolist() == [taskid]
        else:
            with pytest.raises(ValueError):
                s[:taskid]








import os

import numpy as np
from torchvision import datasets as torchdata

from continuum.datasets import ImageFolderDataset
from continuum import download


class DTD(ImageFolderDataset):
    """Describable Textures Dataset (DTD)

    Reference:
        * Describing Textures in the Wild
          M. Cimpoi and S. Maji and I. Kokkinos and S. Mohamed and and A. Vedaldi
          CVPR 2014
    """
    url = "https://www.robots.ox.ac.uk/~vgg/data/dtd/download/dtd-r1.0.1.tar.gz"

    def __init__(self, data_path: str, train: bool = True, download: bool = True, split: int = 1):
        super().__init__(data_path=data_path, train=train, download=download, data_type="image_path")

        if not (1 <= int(split) <= 10):
            raise ValueError(f"Available splits are [1, ..., 10], not {split}")
        self.split = split

    def _download(self):
        archive_path = os.path.join(self.data_path, "dtd-r1.0.1.tar.gz")
        if not os.path.exists(archive_path):
            print("Downloading DTD dataset...")
            download.download(self.url, self.data_path)
        if not os.path.exists(os.path.join(self.data_path, "dtd")):
            print("Uncompressing images...")
            download.untar(archive_path)

    def get_data(self):
        x, y, t = self._format(torchdata.ImageFolder(os.path.join(self.data_path, "dtd", "images")).imgs)

        if self.train:
            index_files = [
                os.path.join(self.data_path, "dtd", "labels", f"train{str(self.split)}.txt"),
                os.path.join(self.data_path, "dtd", "labels", f"val{str(self.split)}.txt")
            ]
        else:
            index_files = [
                os.path.join(self.data_path, "dtd", "labels", f"test{str(self.split)}.txt")
            ]

        valid_paths = set()
        for index_file in index_files:
            with open(index_file) as f:
                valid_paths.update(
                    map(lambda p: os.path.join(self.data_path, "dtd", "images", p.strip()),
                        f.readlines()
                    )
                )
        valid_paths = np.array(list(valid_paths))
        indexes = np.isin(x, valid_paths)

        return x[indexes], y[indexes], None



import os
import glob

import pytest
import numpy as np
from PIL import Image
import torch
from torch.utils.data import DataLoader

from continuum.datasets import InMemoryDataset
from continuum.scenarios import SegmentationClassIncremental


def _clean(pattern):
    for f in glob.glob(pattern):
        os.remove(f)


def create_dataset(tmpdir, prefix, png=False, train=True):
    nb_samples = 20

    x = np.random.randint(0, 255, (nb_samples, 2, 2, 3), dtype=np.uint8)
    y = np.zeros((nb_samples, 2, 2), dtype=np.uint8)
    y[0:15, 0, 0] = 255

    y[0:10, 0, 1] = 1
    y[4:10, 1, 0] = 2
    y[5:20, 0, 1] = 3
    y[15:20, 1, 1] = 4

    x_paths, y_paths = [], []
    for i in range(nb_samples):
        if png:
            x_paths.append(os.path.join(tmpdir, f"{prefix}_{i}.png"))
        else:
            x_paths.append(os.path.join(tmpdir, f"{prefix}_{i}.jpg"))
        y_paths.append(os.path.join(tmpdir, f"{prefix}_{i}.png"))

        Image.fromarray(x[i]).save(x_paths[-1])
        Image.fromarray(y[i]).save(y_paths[-1])

    return InMemoryDataset(
        np.array(x_paths), np.array(y_paths),
        data_type="segmentation",
        train=train
    )


def create_dense_dataset(tmpdir, png=False, train=False):
    prefix = "dense"
    nb_samples = 20

    x = np.random.randint(0, 255, (nb_samples, 3, 3, 3), dtype=np.uint8)
    y = np.zeros((nb_samples, 3, 3), dtype=np.uint8)
    y[:, 0, 0] = 255
    y[:, 0, 1] = 0
    y[:, 0, 2] = 1
    y[:, 1, 0] = 2
    y[:, 2, 0] = 3
    y[:, 1, 1] = 4

    x_paths, y_paths = [], []
    for i in range(nb_samples):
        if png:
            x_paths.append(os.path.join(tmpdir, f"{prefix}_{i}.png"))
        else:
            x_paths.append(os.path.join(tmpdir, f"{prefix}_{i}.jpg"))
        y_paths.append(os.path.join(tmpdir, f"{prefix}_{i}.png"))

        Image.fromarray(x[i]).save(x_paths[-1])
        Image.fromarray(y[i]).save(y_paths[-1])

    return InMemoryDataset(
        np.array(x_paths), np.array(y_paths),
        data_type="segmentation",
        train=train
    )


@pytest.fixture
def dataset(tmpdir):
    return create_dataset(tmpdir, "seg")


@pytest.fixture
def dataset_test(tmpdir):
    return create_dataset(tmpdir, "seg_test", train=False)


@pytest.fixture
def dataset_dense_test(tmpdir):
    return create_dense_dataset(tmpdir, "seg_dense_test", train=False)


@pytest.fixture
def dataset_png(tmpdir):
    return create_dataset(tmpdir, "seg_png", png=True)



@pytest.mark.parametrize("mode,lengths,increment", [
    ("overlap", (10, 15), 2),
    ("disjoint", (5, 15), 2),
    ("disjoint", (5, 10, 5), 1),
    ("overlap", (10, 15, 5), 1),
    ("sequential", (5, 15), 2),
    ("sequential", (5, 10, 5), 1),
])
def test_length_taskset(dataset, mode, lengths, increment):
    scenario = SegmentationClassIncremental(
        dataset,
        nb_classes=4,
        increment=increment,
        initial_increment=2,
        mode=mode
    )

    assert len(scenario) == len(lengths)
    for i, l in enumerate(lengths):
        assert len(scenario[i]) == l, i


def test_save_indexes(tmpdir):
    dataset = create_dataset(tmpdir, "seg_tmp")
    _clean(os.path.join(tmpdir, "seg_tmp*"))

    with pytest.raises(Exception):
        scenario = SegmentationClassIncremental(
            dataset,
            nb_classes=4,
            increment=2,
            mode="overlap",
        )

    dataset = create_dataset(tmpdir, "seg_tmp")
    scenario = SegmentationClassIncremental(
        dataset,
        nb_classes=4,
        increment=2,
        mode="overlap",
        save_indexes=os.path.join(tmpdir, "indexes.npy")
    )
    _clean(os.path.join(tmpdir, "seg_tmp*"))
    scenario = SegmentationClassIncremental(
        dataset,
        nb_classes=4,
        increment=2,
        mode="overlap",
        save_indexes=os.path.join(tmpdir, "indexes.npy")
    )


def test_advanced_indexing_step(dataset):
    scenario = SegmentationClassIncremental(
        dataset,
        nb_classes=4,
        increment=1,
        mode="overlap"
    )

    with pytest.raises(ValueError):
        task_set = scenario[0:4:2]


@pytest.mark.parametrize("mode,start,end,classes,train", [
    ("overlap", 0, 4, [1, 2, 3, 4], False),
    ("overlap", 0, 4, [1, 2, 3, 4], True),
    ("overlap", 3, 4, [4], True),
    ("overlap", 1, 3, [2, 3], True),
    ("disjoint", 0, 4, [1, 2, 3, 4], True),
    ("disjoint", 3, 4, [4], True),
    ("disjoint", 1, 3, [2, 3], True),
    ("sequential", 0, 4, [1, 2, 3, 4], True),
    ("sequential", 3, 4, [3, 4], True),
    ("sequential", 1, 3, [1, 2, 3], True),
])
def test_advanced_indexing(dataset, dataset_test, mode, start, end, classes, train):
    scenario = SegmentationClassIncremental(
        dataset if train else dataset_test,
        nb_classes=4,
        increment=1,
        mode=mode
    )

    task_set = scenario[start:end]
    loader = DataLoader(task_set, batch_size=200, drop_last=False)
    _, y, t = next(iter(loader))

    t = torch.unique(t)
    y = torch.unique(y)

    assert len(t) == 1 and t[0] == end - 1
    assert set(y.numpy().tolist()) - set([0, 255]) == set(classes)



@pytest.mark.parametrize("mode,all_seen_tasks", [
    ("overlap", False),
    ("overlap", True),
    ("disjoint", True),
    ("disjoint", False),
    ("sequential", True),
    ("sequential", False),
])
def test_labels_test(dataset_test, mode, all_seen_tasks):
    scenario = SegmentationClassIncremental(
        dataset_test,
        nb_classes=4,
        increment=1,
        mode=mode
    )
    classes = [0, 255, 1, 2, 3, 4]

    for task_id in range(len(scenario)):
        if all_seen_tasks:
            task_set = scenario[:task_id+1]
        else:
            task_set = scenario[task_id]

        loader = DataLoader(task_set, batch_size=200, drop_last=False)
        x, y, _ = next(iter(loader))
        seen_classes = torch.unique(y)

        inv_subset_classes = classes[task_id + 3:]
        for c in inv_subset_classes:
            assert c not in seen_classes, task_id
        if all_seen_tasks:
            subset_classes = classes[:task_id + 3]
            for c in subset_classes:
                assert c in seen_classes, task_id
        else:
            assert classes[task_id + 2] in seen_classes


def test_labels_overlap_dense_test(dataset_dense_test):
    scenario = SegmentationClassIncremental(
        dataset_dense_test,
        nb_classes=4,
        increment=1,
        mode="overlap"
    )
    classes = [0, 255, 1, 2, 3, 4]

    for task_id, task_set in enumerate(scenario):
        loader = DataLoader(task_set, batch_size=200, drop_last=False)
        x, y, _ = next(iter(loader))
        seen_classes = torch.unique(y)

        subset_classes = classes[:task_id + 3]
        for c in subset_classes:
            assert c in seen_classes, task_id
        inv_subset_classes = classes[task_id + 3:]
        for c in inv_subset_classes:
            assert c not in seen_classes, task_id
        assert len(x) == 20


@pytest.mark.parametrize("mode,test_background,train", [
    ("overlap", True, True),
    ("overlap", False, True),
    ("disjoint", True, True),
    ("disjoint", False, True),
    ("sequential", True, True),
    ("sequential", False, True),
    ("overlap", True, False),
    ("overlap", False, False),
    ("disjoint", True, False),
    ("disjoint", False, False),
    ("sequential", True, False),
    ("sequential", False, False),
])
def test_background_test(dataset, dataset_test, mode, test_background, train):
    scenario = SegmentationClassIncremental(
        dataset if train else dataset_test,
        nb_classes=4,
        increment=2,
        mode=mode,
        test_background=test_background
    )

    for task_set in scenario:
        loader = DataLoader(task_set, batch_size=200, drop_last=False)
        x, y, _ = next(iter(loader))

        if train or test_background:
            assert 0 in y
        else:
            assert 0 not in y


@pytest.mark.parametrize("mode,class_order,error", [
    ("overlap", [1, 2, 3, 4], False),
    ("overlap", [1, 2, 3, 4, 5], True),
    ("overlap", [0, 1, 2, 3], True),
    ("overlap", [1, 2, 3, 255], True),
    ("disjoint", [1, 2, 3, 4], False),
    ("overlap", [4, 3, 2, 1], False),
    ("overlap", [2, 3, 4, 1], False),

])
def test_class_order(dataset_png, mode, class_order, error):
    """We need PNG here because JPG lose some pixels values"""
    increments = [1, 1, 1, 1]

    if error:
        with pytest.raises(ValueError):
            scenario = SegmentationClassIncremental(
                dataset_png,
                nb_classes=4,
                increment=increments,
                class_order=class_order,
                mode=mode
            )
        return
    else:
        scenario = SegmentationClassIncremental(
            dataset_png,
            nb_classes=4,
            increment=increments,
            class_order=class_order,
            mode=mode
        )

    for task_id, task_set in enumerate(scenario):
        loader = DataLoader(task_set, batch_size=200, drop_last=False)
        x, y, _ = next(iter(loader))
        pixels = torch.unique((x * 255).long())

        assert (task_id + 1) in y
        real_class = class_order[task_id]
        assert real_class in pixels, task_id
        original_y = scenario.get_original_targets(y)
        assert real_class in np.unique(original_y), task_id


@pytest.mark.parametrize("mode,increment", [
    ("overlap", 2),
    ("overlap", 1),
    ("disjoint", 2),
    ("disjoint", 1),
    ("sequential", 2),
    ("sequential", 1),
    ("overlap", [2, 2]),
    ("disjoint", [2, 2]),
    ("disjoint", [2, 1, 1]),
    ("overlap", [2, 1, 1]),
    ("disjoint", [1, 3]),
    ("overlap", [1, 3]),
    ("disjoint", [3, 1]),
    ("overlap", [3, 1]),
])
def test_labels(dataset, mode, increment):
    initial_increment = 2
    nb_classes = 4
    min_cls = 1

    scenario = SegmentationClassIncremental(
        dataset,
        nb_classes=nb_classes,
        increment=increment,
        initial_increment=initial_increment,
        mode=mode
    )

    if isinstance(increment, int) and increment == 2:
        increments = [2, 2]
    elif isinstance(increment, int) and increment == 1:
        increments = [2, 1, 1]
    else:
        increments = increment

    for task_id, task_set in enumerate(scenario):
        loader = DataLoader(task_set, batch_size=200, drop_last=False)
        x, y, t = next(iter(loader))

        assert len(x.shape) == 4
        assert len(y.shape) == 3
        assert len(t.shape) == 1
        assert x.shape[2:] == y.shape[1:]

        assert (t == task_id).all()

        seen_classes = set(torch.unique(y).numpy().tolist())

        max_cls = min_cls + increments[task_id]
        assert 0 in seen_classes, task_id
        if 4 not in seen_classes:
            assert 255 in seen_classes, task_id

        for c in list(range(min_cls, nb_classes + min_cls)):
            if mode in ("overlap", "disjoint"):
                if min_cls <= c < max_cls:
                    assert c in seen_classes, (c, task_id, min_cls, max_cls)
                else:
                    assert c not in seen_classes, (c, task_id, min_cls, max_cls)
            elif mode == "sequential":
                if c < max_cls:
                    assert c in seen_classes, (c, task_id, min_cls, max_cls)
                else:
                    assert c not in seen_classes, (c, task_id, min_cls, max_cls)

        min_cls += increments[task_id]



from typing import Tuple

import numpy as np
from torchvision import datasets as torchdata

from continuum.datasets import _ContinuumDataset


class SVHN(_ContinuumDataset):
    def __init__(
            self,
            data_path: str = "",
            train: bool = True,
            download: bool = True
        ):
        super().__init__(data_path=data_path, train=train, download=download)
        self.dataset = torchdata.SVHN(
            self.data_path,
            download=self.download,
            split="train" if train else "test"
        )

    def get_data(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        x, y = np.array(self.dataset.data), np.array(self.dataset.labels)
        x = x.transpose(0, 2, 3, 1)
        return x, y, None



"""This module is mostly a copy/paste from the useful MiB codebase of Cermelli et al.
See there:
- https://github.com/fcdl94/MiB/blob/master/dataset/transform.py
"""

import warnings
import math
import random
import numbers
import collections

import torch
import torchvision.transforms.functional as Fv
import numpy as np
from PIL import Image


_pil_interpolation_to_str = {
    Image.NEAREST: 'PIL.Image.NEAREST',
    Image.BILINEAR: 'PIL.Image.BILINEAR',
    Image.BICUBIC: 'PIL.Image.BICUBIC',
    Image.LANCZOS: 'PIL.Image.LANCZOS',
    Image.HAMMING: 'PIL.Image.HAMMING',
    Image.BOX: 'PIL.Image.BOX',
}


class Compose:
    """Composes several transforms together.

    :param transforms: list of transforms to compose.
    """

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, img, lbl=None):
        if lbl is not None:
            for t in self.transforms:
                img, lbl = t(img, lbl)
            return img, lbl
        else:
            for t in self.transforms:
                img = t(img)
            return img

    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        for t in self.transforms:
            format_string += '\n'
            format_string += '    {0}'.format(t)
        format_string += '\n)'
        return format_string


class Resize:
    """Resize the input PIL Image to the given size.

    :param size: (sequence or int): Desired output size. If size is a sequence like
                 (h, w), output size will be matched to this. If size is an int,
                 smaller edge of the image will be matched to this number.
                 i.e, if height > width, then image will be rescaled to
                 (size * height / width, size)
    :param interpolation: (int, optional): Desired interpolation. Default is
                          ``PIL.Image.BILINEAR``
    """

    def __init__(self, size, interpolation=Image.BILINEAR):
        assert isinstance(size, int) or (isinstance(size, collections.Iterable) and len(size) == 2)
        self.size = size
        self.interpolation = interpolation

    def __call__(self, img, lbl=None):
        """
        :param img: (PIL Image): Image to be scaled.
        :return: PIL Image: Rescaled image.
        """
        if lbl is not None:
            return Fv.resize(img, self.size, self.interpolation), Fv.resize(lbl, self.size, Image.NEAREST)
        else:
            return Fv.resize(img, self.size, self.interpolation)

    def __repr__(self):
        interpolate_str = _pil_interpolation_to_str[self.interpolation]
        return self.__class__.__name__ + '(size={0}, interpolation={1})'.format(self.size, interpolate_str)


class CenterCrop:
    """Crops the given PIL Image at the center.

    :param size: (sequence or int): Desired output size of the crop. If size is an
                 int instead of sequence like (h, w), a square crop (size, size) is
                 made.
    """

    def __init__(self, size):
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size

    def __call__(self, img, lbl=None):
        """
        :param img: (PIL Image): Image to be cropped.
        :return: PIL Image: Cropped image.
        """
        if lbl is not None:
            return Fv.center_crop(img, self.size), Fv.center_crop(lbl, self.size)
        else:
            return Fv.center_crop(img, self.size)

    def __repr__(self):
        return self.__class__.__name__ + f'(size={self.size})'


class Pad:
    """Pad the given PIL Image on all sides with the given "pad" value.

    :param padding: (int or tuple): Padding on each border. If a single int is provided this
                    is used to pad all borders. If tuple of length 2 is provided this is the padding
                    on left/right and top/bottom respectively. If a tuple of length 4 is provided
                    this is the padding for the left, top, right and bottom borders
                    respectively.
    :param fill: (int): Pixel fill value for constant fill. Default is 0.
                 This value is only used when the padding_mode is constant
    :param padding_mode: (str): Type of padding. Should be: constant, edge, reflect or symmetric.
                         Default is constant.
                        - constant: pads with a constant value, this value is specified with fill
                        - edge: pads with the last value at the edge of the image
                        - reflect: pads with reflection of image without repeating the last value on the edge
                            Fvor example, padding [1, 2, 3, 4] with 2 elements on both sides in reflect mode
                            will result in [3, 2, 1, 2, 3, 4, 3, 2]
                        - symmetric: pads with reflection of image repeating the last value on the edge
                            Fvor example, padding [1, 2, 3, 4] with 2 elements on both sides in symmetric mode
                            will result in [2, 1, 1, 2, 3, 4, 4, 3]
    """

    def __init__(self, padding, fill=0, padding_mode='constant'):
        assert isinstance(padding, (numbers.Number, tuple))
        assert isinstance(fill, (numbers.Number, str))
        assert padding_mode in ['constant', 'edge', 'reflect', 'symmetric']
        if isinstance(padding, collections.Sequence) and len(padding) not in [2, 4]:
            raise ValueError("Padding must be an int or a 2, or 4 element tuple, not a " +
                             "{} element tuple".format(len(padding)))

        self.padding = padding
        self.fill = fill
        self.padding_mode = padding_mode

    def __call__(self, img, lbl=None):
        """
        :param img: (PIL Image): Image to be padded.
        :return: PIL Image: Padded image.
        """
        if lbl is not None:
            return Fv.pad(img, self.padding, self.fill, self.padding_mode), Fv.pad(lbl, self.padding, self.fill, self.padding_mode)
        else:
            return Fv.pad(img, self.padding, self.fill, self.padding_mode)

    def __repr__(self):
        return self.__class__.__name__ + '(padding={0}, fill={1}, padding_mode={2})'.\
            format(self.padding, self.fill, self.padding_mode)


class Lambda:
    """Apply a user-defined lambda as a transform.

    :param lambd: (function): Lambda/function to be used for transform.
    """

    def __init__(self, lambd):
        assert callable(lambd), repr(type(lambd).__name__) + " object is not callable"
        self.lambd = lambd

    def __call__(self, img, lbl=None):
        if lbl is not None:
            return self.lambd(img), self.lambd(lbl)
        else:
            return self.lambd(img)

    def __repr__(self):
        return self.__class__.__name__ + '()'


class RandomRotation:
    """Rotate the image by angle.

    :param degrees: (sequence or float or int): Range of degrees to select from.
                    If degrees is a number instead of sequence like (min, max), the range of degrees
                    will be (-degrees, +degrees).
    :param resample: ({PIL.Image.NEAREST, PIL.Image.BILINEAR, PIL.Image.BICUBIC}, optional):
                     An optional resampling filter.
                     See http://pillow.readthedocs.io/en/3.4.x/handbook/concepts.html#filters
                     If omitted, or if the image has mode "1" or "P", it is set to PIL.Image.NEAREST.
    :param expand: (bool, optional): Optional expansion flag.
                   If true, expands the output to make it large enough to hold the entire rotated image.
                   If false or omitted, make the output image the same size as the input image.
                   Note that the expand flag assumes rotation around the center and no translation.
    :param center: (2-tuple, optional): Optional center of rotation.
                   Origin is the upper left corner.
                   Default is the center of the image.
    """

    def __init__(self, degrees, resample=False, expand=False, center=None):
        if isinstance(degrees, numbers.Number):
            if degrees < 0:
                raise ValueError("If degrees is a single number, it must be positive.")
            self.degrees = (-degrees, degrees)
        else:
            if len(degrees) != 2:
                raise ValueError("If degrees is a sequence, it must be of len 2.")
            self.degrees = degrees

        self.resample = resample
        self.expand = expand
        self.center = center

    @staticmethod
    def get_params(degrees):
        """Get parameters for ``rotate`` for a random rotation.

        :return: sequence: params to be passed to ``rotate`` for random rotation.
        """
        angle = random.uniform(degrees[0], degrees[1])

        return angle

    def __call__(self, img, lbl):
        """

        :param img: (PIL Image): Image to be rotated.
        :param lbl: (PIL Image): Label to be rotated.
        :return: PIL Image: Rotated image, PIL Image: Rotated label.
        """

        angle = self.get_params(self.degrees)
        if lbl is not None:
            return Fv.rotate(img, angle, self.resample, self.expand, self.center), \
                   Fv.rotate(lbl, angle, self.resample, self.expand, self.center)
        else:
            return Fv.rotate(img, angle, self.resample, self.expand, self.center)

    def __repr__(self):
        format_string = self.__class__.__name__ + '(degrees={0}'.format(self.degrees)
        format_string += ', resample={0}'.format(self.resample)
        format_string += ', expand={0}'.format(self.expand)
        if self.center is not None:
            format_string += ', center={0}'.format(self.center)
        format_string += ')'
        return format_string


class RandomHorizontalFvlip:
    """Horizontally flip the given PIL Image randomly with a given probability.

    :param p: (float): probability of the image being flipped. Default value is 0.5
    """

    def __init__(self, p=0.5):
        self.p = p

    def __call__(self, img, lbl=None):
        """
        :param img: (PIL Image): Image to be flipped.
        :return: PIL Image: Randomly flipped image.
        """
        if random.random() < self.p:
            if lbl is not None:
                return Fv.hflip(img), Fv.hflip(lbl)
            else:
                return Fv.hflip(img)
        if lbl is not None:
            return img, lbl
        else:
            return img

    def __repr__(self):
        return self.__class__.__name__ + '(p={})'.format(self.p)


class RandomVerticalFvlip:
    """Vertically flip the given PIL Image randomly with a given probability.

    :param p: (float): probability of the image being flipped. Default value is 0.5
    """

    def __init__(self, p=0.5):
        self.p = p

    def __call__(self, img, lbl):
        """
        :param img: (PIL Image): Image to be flipped.
        :param lbl: (PIL Image): Label to be flipped.
        :return: PIL Image: Randomly flipped image, PIL Image: Randomly flipped label.
        """
        if random.random() < self.p:
            if lbl is not None:
                return Fv.vflip(img), Fv.vflip(lbl)
            else:
                return Fv.vflip(img)
        if lbl is not None:
            return img, lbl
        else:
            return img

    def __repr__(self):
        return self.__class__.__name__ + '(p={})'.format(self.p)


class ToTensor:
    """Convert a ``PIL Image`` or ``numpy.ndarray`` to tensor.

    Converts a PIL Image or numpy.ndarray (H x W x C) in the range
    [0, 255] to a torch.FvloatTensor of shape (C x H x W) in the range [0.0, 1.0]
    if the PIL Image belongs to one of the modes (L, LA, P, I, Fv, RGB, YCbCr, RGBA, CMYK, 1)
    or if the numpy.ndarray has dtype = np.uint8
    In the other cases, tensors are returned without scaling.

    """

    def __call__(self, pic, lbl=None):
        """
        Note that labels will not be normalized to [0, 1].

        :param pic: (PIL Image or numpy.ndarray): Image to be converted to tensor.
        :param lbl: (PIL Image or numpy.ndarray): Label to be converted to tensor.
        :return: Tensor: Converted image and label
        """
        if lbl is not None:
            return Fv.to_tensor(pic), torch.from_numpy(np.array(lbl, dtype=np.uint8))
        else:
            return Fv.to_tensor(pic)

    def __repr__(self):
        return self.__class__.__name__ + '()'


class Normalize:
    """Normalize a tensor image with mean and standard deviation.
    Given mean: ``(M1,...,Mn)`` and std: ``(S1,..,Sn)`` for ``n`` channels, this transform
    will normalize each channel of the input ``torch.*Tensor`` i.e.
    ``input[channel] = (input[channel] - mean[channel]) / std[channel]``

    :param mean: (sequence): Sequence of means for each channel.
    :param std: (sequence): Sequence of standard deviations for each channel.
    """

    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, tensor, lbl=None):
        """
        :param tensor: (Tensor): Tensor image of size (C, H, W) to be normalized.
        :param lbl: (Tensor): Tensor of label. A dummy input for ExtCompose
        :return: Tensor: Normalized Tensor image, Tensor: Unchanged Tensor label
        """
        if lbl is not None:
            return Fv.normalize(tensor, self.mean, self.std), lbl
        else:
            return Fv.normalize(tensor, self.mean, self.std)

    def __repr__(self):
        return self.__class__.__name__ + '(mean={0}, std={1})'.format(self.mean, self.std)


class RandomCrop:
    """Crop the given PIL Image at a random location.

    :param size: (sequence or int): Desired output size of the crop. If size is an
                 int instead of sequence like (h, w), a square crop (size, size) is
                 made.
    :param padding: (int or sequence, optional): Optional padding on each border
                    of the image. Default is 0, i.e no padding. If a sequence of length
                    4 is provided, it is used to pad left, top, right, bottom borders
                    respectively.
    :param pad_if_needed: (boolean): It will pad the image if smaller than the
                          desired size to avoid raising an exception.
    """

    def __init__(self, size, padding=0, pad_if_needed=False):
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size
        self.padding = padding
        self.pad_if_needed = pad_if_needed

    @staticmethod
    def get_params(img, output_size):
        """Get parameters for ``crop`` for a random crop.

        :param img: (PIL Image): Image to be cropped.
        :param output_size: (tuple): Expected output size of the crop.
        :return: tuple: params (i, j, h, w) to be passed to ``crop`` for random crop.
        """
        w, h = img.size
        th, tw = output_size
        if w == tw and h == th:
            return 0, 0, h, w

        i = random.randint(0, h - th)
        j = random.randint(0, w - tw)
        return i, j, th, tw

    def __call__(self, img, lbl=None):
        """
        :param img: (PIL Image): Image to be cropped.
        :param lbl: (PIL Image): Label to be cropped.
        :return: PIL Image: Cropped image, PIL Image: Cropped label.
        """
        if lbl is None:
            if self.padding > 0:
                img = Fv.pad(img, self.padding)
            # pad the width if needed
            if self.pad_if_needed and img.size[0] < self.size[1]:
                img = Fv.pad(img, padding=int((1 + self.size[1] - img.size[0]) / 2))
            # pad the height if needed
            if self.pad_if_needed and img.size[1] < self.size[0]:
                img = Fv.pad(img, padding=int((1 + self.size[0] - img.size[1]) / 2))

            i, j, h, w = self.get_params(img, self.size)

            return Fv.crop(img, i, j, h, w)

        else:
            assert img.size == lbl.size, 'size of img and lbl should be the same. %s, %s' % (img.size, lbl.size)
            if self.padding > 0:
                img = Fv.pad(img, self.padding)
                lbl = Fv.pad(lbl, self.padding)

            # pad the width if needed
            if self.pad_if_needed and img.size[0] < self.size[1]:
                img = Fv.pad(img, padding=int((1 + self.size[1] - img.size[0]) / 2))
                lbl = Fv.pad(lbl, padding=int((1 + self.size[1] - lbl.size[0]) / 2))

            # pad the height if needed
            if self.pad_if_needed and img.size[1] < self.size[0]:
                img = Fv.pad(img, padding=int((1 + self.size[0] - img.size[1]) / 2))
                lbl = Fv.pad(lbl, padding=int((1 + self.size[0] - lbl.size[1]) / 2))

            i, j, h, w = self.get_params(img, self.size)

            return Fv.crop(img, i, j, h, w), Fv.crop(lbl, i, j, h, w)

    def __repr__(self):
        return self.__class__.__name__ + '(size={0}, padding={1})'.format(self.size, self.padding)


class RandomResizedCrop:
    """Crop the given PIL Image to random size and aspect ratio.
    A crop of random size (default: of 0.08 to 1.0) of the original size and a random
    aspect ratio (default: of 3/4 to 4/3) of the original aspect ratio is made. This crop
    is finally resized to given size.
    This is popularly used to train the Inception networks.

    :param size: expected output size of each edge
    :param scale: range of size of the origin size cropped
    :param ratio: range of aspect ratio of the origin aspect ratio cropped
    :param interpolation: Default: PIL.Image.BILINEAR
    """

    def __init__(self, size, scale=(0.08, 1.0), ratio=(3. / 4., 4. / 3.), interpolation=Image.BILINEAR):
        if isinstance(size, tuple):
            self.size = size
        else:
            self.size = (size, size)
        if (scale[0] > scale[1]) or (ratio[0] > ratio[1]):
            warnings.warn("range should be of kind (min, max)")

        self.interpolation = interpolation
        self.scale = scale
        self.ratio = ratio

    @staticmethod
    def get_params(img, scale, ratio):
        """Get parameters for ``crop`` for a random sized crop.

        :param img: (PIL Image): Image to be cropped.
        :param scale: (tuple): range of size of the origin size cropped
        :param ratio: (tuple): range of aspect ratio of the origin aspect ratio cropped
        :return tuple: params (i, j, h, w) to be passed to ``crop`` for a random
                       sized crop.
        """
        area = img.size[0] * img.size[1]

        for attempt in range(10):
            target_area = random.uniform(*scale) * area
            log_ratio = (math.log(ratio[0]), math.log(ratio[1]))
            aspect_ratio = math.exp(random.uniform(*log_ratio))

            w = int(round(math.sqrt(target_area * aspect_ratio)))
            h = int(round(math.sqrt(target_area / aspect_ratio)))

            if w <= img.size[0] and h <= img.size[1]:
                i = random.randint(0, img.size[1] - h)
                j = random.randint(0, img.size[0] - w)
                return i, j, h, w

        # Fvallback to central crop
        in_ratio = img.size[0] / img.size[1]
        if (in_ratio < min(ratio)):
            w = img.size[0]
            h = int(round(w / min(ratio)))
        elif (in_ratio > max(ratio)):
            h = img.size[1]
            w = int(round(h * max(ratio)))
        else:  # whole image
            w = img.size[0]
            h = img.size[1]
        i = (img.size[1] - h) // 2
        j = (img.size[0] - w) // 2
        return i, j, h, w

    def __call__(self, img, lbl=None):
        """
        :param img: (PIL Image): Image to be cropped and resized.
        :return: PIL Image: Randomly cropped and resized image.
        """
        i, j, h, w = self.get_params(img, self.scale, self.ratio)
        if lbl is not None:
            return Fv.resized_crop(img, i, j, h, w, self.size, self.interpolation), \
                   Fv.resized_crop(lbl, i, j, h, w, self.size, Image.NEAREST)
        else:
            return Fv.resized_crop(img, i, j, h, w, self.size, self.interpolation)

    def __repr__(self):
        interpolate_str = _pil_interpolation_to_str[self.interpolation]
        format_string = self.__class__.__name__ + '(size={0}'.format(self.size)
        format_string += ', scale={0}'.format(tuple(round(s, 4) for s in self.scale))
        format_string += ', ratio={0}'.format(tuple(round(r, 4) for r in self.ratio))
        format_string += ', interpolation={0})'.format(interpolate_str)
        return format_string


class ColorJitter:
    """Randomly change the brightness, contrast and saturation of an image.

    :param brightness: (float or tuple of float (min, max)): How much to jitter brightness.
                       brightness_factor is chosen uniformly from [max(0, 1 - brightness), 1 + brightness]
                       or the given [min, max]. Should be non negative numbers.
    :param contrast: (float or tuple of float (min, max)): How much to jitter contrast.
                     contrast_factor is chosen uniformly from [max(0, 1 - contrast), 1 + contrast]
                     or the given [min, max]. Should be non negative numbers.
    :param saturation: (float or tuple of float (min, max)): How much to jitter saturation.
                       saturation_factor is chosen uniformly from [max(0, 1 - saturation), 1 + saturation]
                       or the given [min, max]. Should be non negative numbers.
    :param hue: (float or tuple of float (min, max)): How much to jitter hue.
                hue_factor is chosen uniformly from [-hue, hue] or the given [min, max].
                Should have 0<= hue <= 0.5 or -0.5 <= min <= max <= 0.5.
    """
    def __init__(self, brightness=0, contrast=0, saturation=0, hue=0):
        self.brightness = self._check_input(brightness, 'brightness')
        self.contrast = self._check_input(contrast, 'contrast')
        self.saturation = self._check_input(saturation, 'saturation')
        self.hue = self._check_input(hue, 'hue', center=0, bound=(-0.5, 0.5),
                                     clip_first_on_zero=False)

    def _check_input(self, value, name, center=1, bound=(0, float('inf')), clip_first_on_zero=True):
        if isinstance(value, numbers.Number):
            if value < 0:
                raise ValueError("If {} is a single number, it must be non negative.".format(name))
            value = [center - value, center + value]
            if clip_first_on_zero:
                value[0] = max(value[0], 0)
        elif isinstance(value, (tuple, list)) and len(value) == 2:
            if not bound[0] <= value[0] <= value[1] <= bound[1]:
                raise ValueError("{} values should be between {}".format(name, bound))
        else:
            raise TypeError("{} should be a single number or a list/tuple with lenght 2.".format(name))

        # if value is 0 or (1., 1.) for brightness/contrast/saturation
        # or (0., 0.) for hue, do nothing
        if value[0] == value[1] == center:
            value = None
        return value

    @staticmethod
    def get_params(brightness, contrast, saturation, hue):
        """Get a randomized transform to be applied on image.
        Arguments are same as that of __init__.

        :return: Transform which randomly adjusts brightness, contrast and
                 saturation in a random order.
        """
        transforms = []

        if brightness is not None:
            brightness_factor = random.uniform(brightness[0], brightness[1])
            transforms.append(Lambda(lambda img: Fv.adjust_brightness(img, brightness_factor)))

        if contrast is not None:
            contrast_factor = random.uniform(contrast[0], contrast[1])
            transforms.append(Lambda(lambda img: Fv.adjust_contrast(img, contrast_factor)))

        if saturation is not None:
            saturation_factor = random.uniform(saturation[0], saturation[1])
            transforms.append(Lambda(lambda img: Fv.adjust_saturation(img, saturation_factor)))

        if hue is not None:
            hue_factor = random.uniform(hue[0], hue[1])
            transforms.append(Lambda(lambda img: Fv.adjust_hue(img, hue_factor)))

        random.shuffle(transforms)
        transform = Compose(transforms)

        return transform

    def __call__(self, img, lbl=None):
        """
        :param img: (PIL Image): Input image.
        :return: PIL Image: Color jittered image.
        """
        transform = self.get_params(self.brightness, self.contrast,
                                    self.saturation, self.hue)
        if lbl is not None:
            return transform(img), lbl
        else:
            return transform(img)

    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        format_string += 'brightness={0}'.format(self.brightness)
        format_string += ', contrast={0}'.format(self.contrast)
        format_string += ', saturation={0}'.format(self.saturation)
        format_string += ', hue={0})'.format(self.hue)
        return format_string



from typing import Union, Callable, Tuple, Any

import numpy as np

from continuum.rehearsal import herd_random, herd_closest_to_cluster, herd_closest_to_barycenter


class RehearsalMemory:
    """Class to handle the rehearsal data.

    # TODO: how does it handle non-class incremental setting?
    # TODO: what if new data contains old classes?

    :param memory_size: Total amount of memory (in number of samples) to allocate.
    :param herding_method: A string indicating how to herd, or a function. Please
                           see the predfined herding methods if you want to
                           create your own.
    :param fixed_memory: Fix the amount per class instead of using all available
                         memory capacity. This is usually fixed for paper starting
                         with half the classes like LUCIR (Hou et al. CVPR2019)
                         and PODNet (Douillard et al. ECCV2020).
    :param nb_total_classes: In case of fixed memory, precise the total amount of
                             classes that will be seen.
    """
    def __init__(
        self,
        memory_size: int,
        herding_method: Union[str, Callable],
        fixed_memory: bool = False,
        nb_total_classes: Union[None, int] = None
    ):
        if isinstance(herding_method, str):
            if herding_method == "random":
                herding_method = herd_random
            elif herding_method == "cluster":
                herding_method = herd_closest_to_cluster
            elif herding_method == "barycenter":
                herding_method = herd_closest_to_barycenter
            else:
                raise NotImplementedError(
                    f"Unknown rehearsal method {herding_method}, "
                    f"Provide a string ('random', 'cluster', or 'barycenter')"
                )
        elif hasattr(herding_method, "__call__"):
            pass
        else:
            raise NotImplementedError(
               f"Unknown rehearsal method {herding_method}, "
               f"Either provide its string name or a callable."
            )

        if fixed_memory and nb_total_classes is None:
            raise ValueError(
                f"When memory is fixed per class, you need to provide the `nb_total_classes` info."
            )

        self.memory_size = memory_size
        self.herding_method = herding_method
        self.fixed_memory = fixed_memory
        self.nb_total_classes = nb_total_classes
        self.seen_classes = set()

        self._x = self._y = self._t = None

    @property
    def nb_classes(self) -> int:
        """Current number of seen classes."""
        return len(self.seen_classes)

    @property
    def memory_per_class(self) -> int:
        """Number of samples per classes."""
        if self.fixed_memory:
            return self.memory_size // self.nb_total_classes
        elif self.nb_classes > 0:
            return self.memory_size // self.nb_classes
        return self.memory_size

    def get(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Get the memory data to be added to a TaskSet."""
        if self._x is None:
            raise ValueError("Memory is empty!")
        return self._x, self._y, self._t

    def __len__(self) -> int:
        """Amount of samples stored in memory."""
        if self._x is None:
            return 0
        return len(self._x)

    def _reduce(self) -> None:
        """Reduce the amount of old classes when new classes are stored in memory."""
        x, y, t = [], [], []
        for class_id in np.unique(self._y):
            indexes = np.where(self._y == class_id)[0]
            if len(indexes) > self.memory_per_class:
                x.append(self._x[indexes[:self.memory_per_class]])
                y.append(self._y[indexes[:self.memory_per_class]])
                t.append(self._t[indexes[:self.memory_per_class]])

        if len(x) > 0:
            self._x = np.concatenate(x)
            self._y = np.concatenate(y)
            self._t = np.concatenate(t)

    def add(
        self,
        x: np.ndarray,
        y: np.ndarray,
        t: np.ndarray,
        z: Any
    ) -> None:
        """Add new classes to the memory.

        :param x: Input data (images, paths, etc.)
        :param y: Labels of the data.
        :param t: Task ids of the data.
        :param z: Extra info, could be many things but usually the extracted features.
        """
        for c in np.unique(y):
            self.seen_classes.add(c)

        mem_x, mem_y, mem_t = self.herding_method(x, y, t, z, nb_per_class=self.memory_per_class)

        if self._x is None:
            self._x, self._y, self._t = mem_x, mem_y, mem_t
        else:
            if not self.fixed_memory:
                self._reduce()
            self._x = np.concatenate((self._x, mem_x))
            self._y = np.concatenate((self._y, mem_y))
            self._t = np.concatenate((self._t, mem_t))



import os

import pytest
import numpy as np

from continuum import ContinualScenario
from continuum.datasets import CTRLplus, CTRLminus, CTRLout, CTRLin, CTRLplastic


DATA_PATH = os.environ.get("CONTINUUM_DATA_PATH")


@pytest.mark.slow
@pytest.mark.parametrize("dataset,nb_tasks,classes_per_task,qt_train,qt_val", [
    (CTRLminus, 6, [
        list(range(10)), list(range(10, 20)), list(range(20, 67)),
        list(range(67, 77)), list(range(77, 87)), list(range(10))],
        [4000, 400, 376, 400, 400, 400],
        [2000, 200, 188, 200, 200, 200]
    ),
    (CTRLplus, 6, [
        list(range(10)), list(range(10, 20)), list(range(20, 67)),
        list(range(67, 77)), list(range(77, 87)), list(range(10))],
        [400, 400, 376, 400, 400, 4000],
        [200, 200, 188, 200, 200, 2000]
    ),
    (CTRLin, 6, [
        list(range(10)), list(range(10, 20)), list(range(20, 67)),
        list(range(67, 77)), list(range(77, 87)), list(range(10))],
        [4000, 400, 376, 400, 400, 50],
        [2000, 200, 188, 200, 200, 30]
    ),
    (CTRLout, 6, [
        list(range(10)), list(range(10, 20)), list(range(20, 67)),
        list(range(67, 77)), list(range(77, 87)), list(range(87, 97))],
        [4000, 400, 376, 400, 400, 400],
        [2000, 200, 188, 200, 200, 200]
    ),
    (CTRLplastic, 5, [
        list(range(10)), list(range(10, 57)),
        list(range(57, 67)), list(range(67, 77)), list(range(77, 87))],
        [400, 376, 400, 400, 4000],
        [200, 188, 200, 200, 2000]
    ),
])
def test_ctrl(tmpdir, dataset, nb_tasks, classes_per_task, qt_train, qt_val):
    path = DATA_PATH or tmpdir  # Use env variable else pytest default temp dir

    s_train = ContinualScenario(dataset(path, split="train", download=True))
    s_val = ContinualScenario(dataset(path, split="val", download=True))
    s_test = ContinualScenario(dataset(path, split="test", download=True))

    assert len(s_train) == len(s_val) == len(s_test) == nb_tasks

    for i, (tr_set, va_set, te_set) in enumerate(zip(s_train, s_val, s_test)):
        assert np.unique(tr_set._y).tolist() == np.unique(va_set._y).tolist() == np.unique(te_set._y).tolist()
        assert np.unique(tr_set._y).tolist() == classes_per_task[i], i
        assert len(tr_set) == qt_train[i]
        assert len(va_set) == qt_val[i]



import numpy as np
import torch
from continuum.datasets import InMemoryDataset
from continuum.scenarios import ContinualScenario


def create_subscenario(base_scenario, task_indexes):
    """
    In this function we want to create a subscenario from the different tasks, either by subsampling tasks or reodering
    or both.
    """

    if torch.is_tensor(task_indexes):
        task_indexes = task_indexes.numpy()

    new_x, new_y, new_t = None, None, None
    if base_scenario.cl_dataset.bounding_boxes is not None:
        raise ValueError("the function create_subscenario is not compatible with scenario with bounding_boxes yet.")

    for i, index in enumerate(task_indexes):
        taskset = base_scenario[index]
        all_task_indexes = np.arange(len(taskset))
        x, y, t = taskset.get_raw_samples(all_task_indexes)
        t = np.ones(len(y)) * i
        if new_x is None:
            new_x = x
            new_y = y
            new_t = t
        else:
            new_x = np.concatenate([new_x, x], axis=0)
            new_y = np.concatenate([new_y, y], axis=0)
            new_t = np.concatenate([new_t, t], axis=0)
    dataset = InMemoryDataset(new_x, new_y, new_t, data_type=base_scenario.cl_dataset.data_type)

    return ContinualScenario(dataset, transformations=base_scenario.transformations)



# pylint: disable=C0401
# flake8: noqa
from continuum.generators.scenarios_generator import TaskOrderGenerator, ClassOrderGenerator, HashGenerator

__all__ = [
    "TaskOrderGenerator",
    "ClassOrderGenerator",
    "HashGenerator"
]



import warnings
from copy import copy
from typing import Callable, List, Union, Optional
import os
import multiprocessing

import numpy as np
from PIL import Image
import torchvision

from continuum.datasets import _ContinuumDataset
from continuum.scenarios import ClassIncremental
from continuum.tasks import TaskSet
from continuum.download import ProgressBar


class SegmentationClassIncremental(ClassIncremental):
    """Continual Loader, generating datasets for the consecutive tasks.

    References
        * Incremental Learning Techniques for Semantic Segmentation
          Umberto Michieli, Pietro Zanuttigh
          ICCV Workshop 2017
        * Modeling the Background for Incremental Learning in Semantic Segmentation
          Fabio Cermelli, Massimiliano Mancini, Samuel Rota Bulò, Elisa Ricci, Barbara Caputo
          CVPR 2020
        * PLOP: Learning without Forgetting for Continual Semantic Segmentation
          Arthur Douillard, Yifu Chen, Arnaud Dapogny, Matthieu Cord
          CVPR 2021

    :param cl_dataset: A continual dataset.
    :param nb_classes: The number of classes of the dataset (excluding bg=0 and unk=255).
    :param increment: Either number of classes per task (e.g. increment=2),
                    or a list specifying for every task the amount of new classes
                     (e.g. increment=[5,1,1,1,1]).
    :param initial_increment: A different task size applied only for the first task.
                              Desactivated if `increment` is a list.
    :param transformations: A list of transformations applied to all tasks.
    :param class_order: An optional custom class order, used for NC.
                        e.g. [0,1,2,3,4,5,6,7,8,9] or [5,2,4,1,8,6,7,9,0,3]
    :param mode: The mode of incremental segmentation. In both "sequential" and
                 "disjoint", images only contain pixels of old or current classes,
                 while in "overlap", future classes can also be present.
                 In "sequential" all pixels are labelized while in "disjoint" and
                 "overlap" only pixels of the current classes are labelized.
    :param save_indexes: Path where to save and load the indexes of the different
                         tasks. Computing it may be slow, so it can be worth to
                         checkpoint those.
    :param test_background: Whether to ignore the background (0) during the testing
                            phase (False) or to keep its label (True).
    """

    def __init__(
        self,
        cl_dataset: _ContinuumDataset,
        nb_classes: int,
        increment: Union[List[int], int] = 0,
        initial_increment: int = 0,
        transformations: List[Callable] = None,
        class_order: Optional[List[int]] = None,
        mode: str = "overlap",
        save_indexes: Optional[str] = None,
        test_background: bool = True
    ) -> None:
        self.mode = mode
        self.save_indexes = save_indexes
        self.test_background = test_background
        self._nb_classes = nb_classes

        if cl_dataset.data_type != "segmentation":
            raise ValueError(
                f"Dataset {cl_dataset} doesn't have the right data type but "
                f"{self.cl_dataset.data_type}."
            )

        if self.mode not in ("overlap", "disjoint", "sequential"):
            raise ValueError(f"Unknown mode={mode}.")

        if class_order is not None:
            if 0 in class_order:
                raise ValueError("Exclude the background (0) from the class order.")
            if 255 in class_order:
                raise ValueError("Exclude the unknown (255) from the class order.")
            if len(class_order) != nb_classes:
                raise ValueError(
                    f"Number of classes ({nb_classes}) != class ordering size ({len(class_order)}."
                )

        super().__init__(
            cl_dataset=cl_dataset,
            increment=increment,
            initial_increment=initial_increment,
            class_order=class_order,
            transformations=transformations,
        )

    @property
    def nb_classes(self) -> int:
        """Total number of classes in the whole continual setting."""
        return self._nb_classes

    def __getitem__(self, task_index: Union[int, slice]) -> TaskSet:
        """Returns a task by its unique index.

        :param task_index: The unique index of a task. As for List, you can use
                           indexing between [0, len], negative indexing, or
                           even slices.
        :return: A train PyTorch's Datasets.
        """
        if isinstance(task_index, slice) and task_index.step is not None:
            raise ValueError("Step in slice for segmentation is not supported.")

        x, y, t, task_index = self._select_data_by_task(task_index)
        t = self._get_task_ids(t, task_index)

        return TaskSet(
            x, y, t,
            self.trsf,
            target_trsf=self._get_label_transformation(task_index),
            data_type=self.cl_dataset.data_type
        )

    def get_original_targets(self, targets: np.ndarray) -> np.ndarray:
        """Returns the original targets not changed by the custom class order.

        :param targets: An array of targets, as provided by the task datasets.
        :return: An array of targets, with their original values.
        """
        return self._class_mapping(targets)

    def _get_label_transformation(self, task_index: Union[int, List[int]]):
        """Returns the transformation to apply on the GT segmentation maps.

        :param task_index: The selected task id.
        :return: A pytorch transformation.
        """
        if isinstance(task_index, int):
            task_index = [task_index]
        if not self.train:
            # In testing mode, all labels brought by previous tasks are revealed
            task_index = list(range(max(task_index) + 1))

        if self.mode in ("overlap", "disjoint"):
            # Previous and future (for disjoint) classes are hidden
            labels = self._get_task_labels(task_index)
        elif self.mode == "sequential":
            # Previous classes are not hidden, no future classes are present
            labels = self._get_task_labels(list(range(max(task_index) + 1)))
        else:
            raise ValueError(f"Unknown mode={self.mode}.")

        inverted_order = {label: self.class_order.index(label) + 1 for label in labels}
        inverted_order[255] = 255

        masking_value = 0
        if not self.train:
            if self.test_background:
                inverted_order[0] = 0
            else:
                masking_value = 255

        return torchvision.transforms.Lambda(
            lambda seg_map: seg_map.apply_(
                lambda v: inverted_order.get(v, masking_value)
            )
        )

    def _get_task_ids(self, t: np.ndarray, task_indexes: Union[int, List[int]]) -> np.ndarray:
        """Reduce multiple task ids to a single one per sample.

        In segmentation, the same image can have several task ids. We assume that
        the selected task ids is the last one as it will be the one that matter
        for the ground-truth segmentation maps.

        :param t: A matrix of task ids of shape (nb_samples, nb_tasks).
        :param task_indexes: The selected task ids.
        :return: An array of task ids of shape (nb_samples,).
        """
        if isinstance(task_indexes, list):
            task_indexes = max(task_indexes)
        return np.ones((len(t))) * task_indexes

    def _get_task_labels(self, task_indexes: Union[int, List[int]]) -> List[int]:
        """Returns the labels/classes of the current tasks, not all present labels!

        :param task_indexes: The selected task ids.
        :return: A list of class/labels ids.
        """
        if isinstance(task_indexes, int):
            task_indexes = [task_indexes]

        labels = set()
        for t in task_indexes:
           previous_inc = sum(self._increments[:t])
           labels.update(
               self.class_order[previous_inc:previous_inc+self._increments[t]]
           )

        return list(labels)

    def _setup(self, nb_tasks: int) -> int:
        """Setups the different tasks."""
        x, y, _ = self.cl_dataset.get_data()
        self.class_order = self.class_order or self.cl_dataset.class_order or list(
            range(1, self._nb_classes + 1))

        # For when the class ordering is changed,
        # so we can quickly find the original labels
        def class_mapping(c):
            if c in (0, 255): return c
            return self.class_order[c - 1]
        self._class_mapping = np.vectorize(class_mapping)

        self._increments = self._define_increments(
            self.increment, self.initial_increment, self.class_order
        )

        # Checkpointing the indexes if the option is enabled.
        # The filtering can take multiple minutes, thus saving/loading them can
        # be useful.
        if self.save_indexes is not None and os.path.exists(self.save_indexes):
            print(f"Loading previously saved indexes ({self.save_indexes}).")
            t = np.load(self.save_indexes)
        else:
            print("Computing indexes, it may be slow!")
            t = _filter_images(
                y, self._increments, self.class_order, self.mode
            )
            if self.save_indexes is not None:
                np.save(self.save_indexes, t)

        assert len(x) == len(y) == len(t) and len(t) > 0

        self.dataset = (x, y, t)

        return len(self._increments)


def _filter_images(
    paths: Union[np.ndarray, List[str]],
    increments: List[int],
    class_order: List[int],
    mode: str = "overlap"
) -> np.ndarray:
    """Select images corresponding to the labels.

    Strongly inspired from Cermelli's code:
    https://github.com/fcdl94/MiB/blob/master/dataset/utils.py#L19

    :param paths: An iterable of paths to gt maps.
    :param increments: All individual increments.
    :param class_order: The class ordering, which may not be [1, 2, ...]. The
                        background class (0) and unknown class (255) aren't
                        in this class order.
    :param mode: Mode of the segmentation (see scenario doc).
    :return: A binary matrix representing the task ids of shape (nb_samples, nb_tasks).
    """
    indexes_to_classes = []
    pb = ProgressBar()

    with multiprocessing.Pool(min(8, multiprocessing.cpu_count())) as pool:
        for i, classes in enumerate(pool.imap(_find_classes, paths), start=1):
            indexes_to_classes.append(classes)
            if i % 100 == 0:
                pb.update(None, 100, len(paths))
        pb.end(len(paths))

    t = np.zeros((len(paths), len(increments)))
    accumulated_inc = 0

    for task_id, inc in enumerate(increments):
        labels = class_order[accumulated_inc:accumulated_inc+inc]
        old_labels = class_order[:accumulated_inc]
        all_labels = labels + old_labels + [0, 255]

        for index, classes in enumerate(indexes_to_classes):
            if mode == "overlap":
                if any(c in labels for c in classes):
                    t[index, task_id] = 1
            elif mode in ("disjoint", "sequential"):
                if any(c in labels for c in classes) and all(c in all_labels for c in classes):
                    t[index, task_id] = 1
            else:
                raise ValueError(f"Unknown mode={mode}.")

        accumulated_inc += inc

    return t


def _find_classes(path: str) -> np.ndarray:
    """Open a ground-truth segmentation map image and returns all unique classes
    contained.

    :param path: Path to the image.
    :return: Unique classes.
    """
    return np.unique(np.array(Image.open(path)).reshape(-1))



import numpy as np
import pytest

from continuum.datasets import InMemoryDataset
from continuum.scenarios import ContinualScenario

# yapf: disable

def test_scenario():
    x = np.ones((100, 4, 4, 3), dtype=np.uint8)
    y = np.arange(100) // 5

    nb_tasks = 10
    t = np.random.randint(nb_tasks, size=100)

    dummy = InMemoryDataset(x, y, t)
    scenario = ContinualScenario(dummy)

    assert scenario.nb_tasks == nb_tasks


def test_bad_task_ids():
    x = np.ones((100, 4, 4, 3), dtype=np.uint8)
    y = np.arange(100) // 5
    nb_tasks = 10
    # test if one missing generate an error
    t = np.random.randint(10, size=100)
    t = t + np.ones(100)  # shift indexes from [0 - 9] to [1 - 10]

    dummy = InMemoryDataset(x, y, t)

    with pytest.raises(Exception):
        scenario = ContinualScenario(dummy)



import os
from typing import Iterable, Set, Tuple, Union

import numpy as np

from continuum import download
from continuum.datasets.base import _ContinuumDataset


class Core50(_ContinuumDataset):
    """Continuum version of the Core50 dataset.

    References:
        * Core50: a new Dataset and Benchmark for Continuous Object Recognition
          Lomonaco & Maltoni.
          CoRL 2017

    :param data_path: The folder path containing the data.
    :param train_image_ids: The image ids belonging to the train set. Either the
                            csv file containing them auto-downloaded, or a list
                            of string.
    :param download: An option useless in this case.
    """

    data_url = "http://bias.csr.unibo.it/maltoni/download/core50/core50_128x128.zip"
    train_ids_url = "https://vlomonaco.github.io/core50/data/core50_train.csv"

    def __init__(
            self,
            data_path: str,
            train: bool = True,
            train_image_ids: Union[str, Iterable[str], None] = None,
            scenario: str = "classes",
            download: bool = True,
            classification: str = "object"
    ):
        assert scenario in ["classes", "domains", "objects"]
        assert classification in ["object", "category"]
        self.train_image_ids = train_image_ids
        super().__init__(data_path=data_path, train=train, download=download)

        self.scenario = scenario
        self.classification = classification

        if self.train_image_ids is None:
            self.train_image_ids = os.path.join(self.data_path, "core50_train.csv")

        if isinstance(self.train_image_ids, str):
            self.train_image_ids = self._read_csv(self.train_image_ids)
        elif isinstance(self.train_image_ids, list):
            self.train_image_ids = set(self.train_image_ids)

    @property
    def data_type(self):
        return "image_path"

    def _download(self):
        if os.path.exists(os.path.join(self.data_path, "core50_128x128")):
            print("Dataset already extracted.")
        else:
            path = download.download(self.data_url, self.data_path)
            download.unzip(path)
            print("Dataset extracted.")

        split_path = os.path.join(self.data_path, "core50_train.csv")
        if self.train_image_ids is None and os.path.exists(split_path):
            self.train_image_ids = split_path
            print("Train/split already downloaded.")
        elif self.train_image_ids is None:
            print("Downloading train/test split.")
            self.train_image_ids = download.download(self.train_ids_url, self.data_path)

    def _read_csv(self, csv_file: str) -> Set[str]:
        """Read the csv file containing the ids of training samples."""
        train_images_ids = set()

        with open(csv_file, "r") as f:
            print(csv_file)
            for line in f:
                image_id = line.split(",")[0].split(".")[0]
                train_images_ids.add(image_id)

        return train_images_ids

    def get_data(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Generate the CORe50 data.

        CORe50, in one of its many iterations, is made of 50 objects, each present
        in 10 different domains (in-door, street, garden, etc.).
        # [o1, ..., o5] -> plug adapters  -> label 1
        # [o6, ..., o10] -> mobile phones
        # [o11, ..., o15] -> scissors
        # [o16, ..., o20] -> light bulbs
        # [o21, ..., o25] -> cans
        # [o26, ..., o30] -> glasses
        # [o31, ..., o35] -> balls
        # [o36, ..., o40] -> markers
        # [o41, ..., o45] -> cups
        # [o46, ..., o50] -> remote controls

        In class incremental (NC) setting, those domains won't matter.

        In instance incremental (NI) setting, the domains come one after the other,
        but all classes are present since the first task. Seven domains are allocated
        for the train set, while 3 domains are allocated for the test set.

        In the case of the test set, all domains have the "dummy" label of 0. The
        authors designed this dataset with a fixed test dataset in mind.
        """
        x, y, t = [], [], []

        domain_counter = 0
        for domain_id in range(11):
            # We walk through the 11 available domains.
            domain_folder = os.path.join(self.data_path, "core50_128x128", f"s{domain_id + 1}")

            has_images = False
            for object_id in range(50):
                # We walk through the 50 available object categories.
                object_folder = os.path.join(domain_folder, f"o{object_id + 1}")

                for path in os.listdir(object_folder):
                    image_id = path.split(".")[0]

                    if (
                            (self.train and image_id not in self.train_image_ids) or  # type: ignore
                            (not self.train and image_id in self.train_image_ids)  # type: ignore
                    ):
                        continue

                    x.append(os.path.join(object_folder, path))
                    if self.classification == "object":
                        # Ranges from 0-49
                        class_label = object_id
                    else:
                        # Ranges from 0-9
                        class_label = object_id // 5
                    y.append(class_label)

                    if self.train:  # We add a new domain id for the train set.
                        if self.scenario == "domains":
                            t.append(domain_counter)
                        elif self.scenario == "classes":
                            t.append(class_label)
                        elif self.scenario == "objects":
                            t.append(object_id)

                    else:  # Test set is fixed, therefore we artificially give a unique domain.
                        t.append(0)

                    has_images = True
            if has_images:
                domain_counter += 1

        x = np.array(x)
        y = np.array(y)
        t = np.array(t)

        return x, y, t


class Core50v2_79(_ContinuumDataset):
    data_url = "http://bias.csr.unibo.it/maltoni/download/core50/core50_128x128.zip"
    splits_url = "https://vlomonaco.github.io/core50/data/batches_filelists_NICv2.zip"
    nb_tasks = 79

    @property
    def data_type(self):
        return "image_path"

    def __init__(self, data_path: str, train: bool = True, download: bool = True, run_id: int = 0):
        if run_id > 9 or run_id < 0:
            raise ValueError(
                "CORe50 v2 only provides split for 10 runs (ids 0 to 9),"
                f" invalid run_id={run_id}."
            )
        self.run_id = run_id

        super().__init__(data_path=data_path, train=train, download=download)

    def _download(self):
        if os.path.exists(os.path.join(self.data_path, "core50_128x128")):
            print("Dataset already extracted.")
        else:
            path = download.download(self.data_url, self.data_path)
            download.unzip(path)
            print("Dataset extracted.")

        if os.path.exists(os.path.join(self.data_path, "batches_filelists_NICv2.zip")):
            print("Split info already downloaded.")
        else:
            path = download.download(self.splits_url, self.data_path)
            download.unzip(path)
            print("Split info extracted.")

    def get_data(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        if self.train:
            return self._train_init()
        return self._test_init()

    def _test_init(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        text_file = os.path.join(
            self.data_path, f"NIC_v2_{self.nb_tasks}", f"run{self.run_id}", "test_filelist.txt"
        )
        paths, targets = self._read_txt(text_file)

        return paths, targets, np.zeros(len(targets))

    def _train_init(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        template = os.path.join(
            self.data_path, f"NIC_v2_{self.nb_tasks}", f"run{self.run_id}",
            "train_batch_{}_filelist.txt"
        )

        paths, targets, tasks = [], [], []
        for task_id in range(self.nb_tasks):
            p, t = self._read_txt(template.format(str(task_id).rjust(2, "0")))

            paths.append(p)
            targets.append(t)
            tasks.append(task_id * np.ones(len(t)))

        paths = np.concatenate(paths)
        targets = np.concatenate(targets)
        tasks = np.concatenate(tasks)

        return paths, targets, tasks

    def _read_txt(self, path: str) -> Tuple[np.ndarray, np.ndarray]:
        """Read Core50 v2 split info that are stored in a txt file.

        The format, for each line, is "path<space>class_id".

        :param path: The path to the text file.
        :return: An array of image paths and an array of targets.
        """
        image_root_path = os.path.join(self.data_path, "core50_128x128")

        paths, targets = [], []
        with open(path, "r") as f:
            for line in f:
                p, t = line.strip().split(" ")
                paths.append(os.path.join(image_root_path, p))
                targets.append(int(t))

        return np.array(paths), np.array(targets)


class Core50v2_196(Core50v2_79):
    nb_tasks = 196


class Core50v2_391(Core50v2_79):
    nb_tasks = 391



import os

import pytest
from torchvision.transforms import Resize, ToTensor

from continuum.datasets import CIFAR100
from continuum.scenarios import ClassIncremental, ContinualScenario

DATA_PATH = os.environ.get("CONTINUUM_DATA_PATH")

'''
Basic test of CIFAR100 without parameters
'''
@pytest.mark.slow
def test_scenario_CIFAR100_ClassIncremental():
    dataset = CIFAR100(DATA_PATH, train=True)
    scenario = ClassIncremental(dataset, increment=50)

    assert scenario.nb_classes == 100
    assert scenario.nb_tasks == 2

'''
Basic test of CIFAR100 without parameters
'''
@pytest.mark.slow
def test_scenario_CIFAR100_CoarseLabels():
    dataset = CIFAR100(DATA_PATH, train=True, labels_type="category")
    scenario = ClassIncremental(dataset, increment=10)

    assert scenario.nb_classes == 20
    assert scenario.nb_tasks == 2

'''
Basic test of CIFAR100 without parameters
'''
@pytest.mark.slow
def test_scenario_CIFAR100_Scenarios():
    dataset = CIFAR100(DATA_PATH, train=True, labels_type="category", task_labels="category")
    scenario = ContinualScenario(dataset)
    assert scenario.nb_classes == 20
    assert scenario.nb_tasks == 20

    dataset = CIFAR100(DATA_PATH, train=True, labels_type="category", task_labels="class")
    scenario = ContinualScenario(dataset)
    assert scenario.nb_classes == 20
    assert scenario.nb_tasks == 100

    dataset = CIFAR100(DATA_PATH, train=True, labels_type="class", task_labels="class")
    scenario = ContinualScenario(dataset)
    assert scenario.nb_classes == 100
    assert scenario.nb_tasks == 100

    dataset = CIFAR100(DATA_PATH, train=True, labels_type="class", task_labels="category")
    scenario = ContinualScenario(dataset)
    assert scenario.nb_classes == 100
    assert scenario.nb_tasks == 20



import os
import zipfile
import tarfile
from six.moves import urllib

import requests


# Fix error when downloading dataset from torchvision
# https://github.com/pytorch/vision/issues/1938#issuecomment-790730080
opener = urllib.request.build_opener()
opener.addheaders = [('User-agent', 'Mozilla/5.0')]
urllib.request.install_opener(opener)


def download(url, path):
    if not os.path.exists(path):
        os.makedirs(path)

    file_name = os.path.join(path, url.split("/")[-1])

    if os.path.exists(file_name):
        print(f"Dataset already downloaded at {file_name}.")
    else:
        urllib.request.urlretrieve(url, file_name, ProgressBar().update)

    return file_name


def unzip(path):
    directory_path = os.path.dirname(path)
    with zipfile.ZipFile(path, 'r') as zip_file:
        zip_file.extractall(directory_path)


def untar(path):
    directory_path = os.path.dirname(path)
    with tarfile.open(path) as tar_file:
        tar_file.extractall(directory_path)


def download_file_from_google_drive(id, destination):
    """Taken from https://stackoverflow.com/a/39225272"""
    URL = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(URL, params = { 'id' : id }, stream = True)
    token = None
    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            token = value
            break

    if token:
        params = { 'id' : id, 'confirm' : token }
        response = session.get(URL, params = params, stream = True)

    CHUNK_SIZE = 32768

    with open(destination, "wb") as f:
        for chunk in response.iter_content(CHUNK_SIZE):
            if chunk: # filter out keep-alive new chunks
                f.write(chunk)


class ProgressBar:
    """Basic Progress Bar.

    Inspired from: https://stackoverflow.com/questions/3173320/text-progress-bar-in-the-console
    """

    def __init__(self):
        self.count = 0

    def update(self, tmp, block_size, total_size):
        self.count += block_size

        percent = f"{int(100 * self.count / total_size)}"
        filled_length = int(100 * self.count // total_size)
        pbar = "#" * filled_length + '-' * (100 - filled_length)

        print("\r|%s| %s%%" % (pbar, percent), end="\r")
        if self.count == total_size:
            print()

    def end(self, total_size):
        self.count = total_size
        self.update(None, 0, total_size)



import os

import numpy as np
import pytest
import torch
from torchvision.transforms import transforms

from continuum.scenarios import Permutations
from tests.test_classorder import InMemoryDatasetTest
from continuum.datasets import MNIST, CIFAR100


DATA_PATH = os.environ.get("CONTINUUM_DATA_PATH")


@pytest.fixture
def numpy_data():
    nb_classes = 6
    nb_data = 100

    x_train = []
    y_train = []
    x_train.append(
        np.array([np.random.randint(100, size=(2, 2, 3)).astype(dtype=np.uint8)] * nb_data)
    )
    y_train.append(np.random.randint(nb_classes, size=(nb_data)))
    x_train = np.concatenate(x_train)
    y_train = np.concatenate(y_train)

    x_test = np.copy(x_train)
    y_test = np.copy(y_train)

    return (x_train, y_train.astype(int)), (x_test, y_test.astype(int))


'''
Test the initialization with three tasks
'''


@pytest.mark.parametrize("seed", [0, 42, 1664, [0, 1, 2], [0, 1, 2, 3]])
def test_init(numpy_data, seed):
    train, test = numpy_data
    dummy = InMemoryDatasetTest(*train)

    nb_tasks = 3
    if isinstance(seed, list):
        nb_tasks = len(seed) + 1

    scenario_1 = Permutations(cl_dataset=dummy, nb_tasks=nb_tasks, seed=seed)
    scenario_2 = Permutations(cl_dataset=dummy, nb_tasks=nb_tasks, seed=seed)

    previous_x = []
    if isinstance(seed, list):
        assert len(scenario_1) == len(scenario_2) == len(seed) + 1

    for task_id, (train_taskset_1, train_taskset_2) in enumerate(zip(scenario_1, scenario_2)):
        assert task_id < nb_tasks

        assert len(train_taskset_1) == len(train_taskset_2)
        indexes = list(range(len(train_taskset_1)))

        x_1, y_1, t_1 = train_taskset_1.get_samples(indexes)
        x_2, y_2, t_2 = train_taskset_2.get_samples(indexes)

        assert (x_1 == x_2).all()
        assert (y_1 == y_2).all()
        assert (t_1 == t_2).all()

        for x in previous_x:
            assert not (x == x_1).all()
        previous_x.append(x_1.clone())


@pytest.mark.slow
@pytest.mark.parametrize("shared_label_space", [True, False])
@pytest.mark.parametrize("dataset", [MNIST, CIFAR100])
def test_with_dataset(dataset, shared_label_space):
    dataset = dataset(data_path=DATA_PATH, download=True, train=True)
    scenario = Permutations(cl_dataset=dataset, nb_tasks=5, seed=0, shared_label_space=shared_label_space)

    for task_id, taskset in enumerate(scenario):

        classes = taskset.get_classes()

        if shared_label_space:
            assert len(classes) == classes.max() + 1
        else:
            assert len(classes) == classes.max() + 1 - (task_id * len(classes))



import numpy as np
import pytest
from torch.utils.data import DataLoader

from continuum.tasks import TaskSet, split_train_val, concat


@pytest.mark.parametrize("nb_others", [1, 2])
def test_concat_function(nb_others):
    x = np.random.rand(10, 2, 2, 3)
    y = np.ones((10,))
    t = np.ones((10,))

    task_sets = [
        TaskSet(np.copy(x), np.copy(y), np.copy(t), None) for _ in range(nb_others)
    ]

    concatenation = concat(task_sets)
    assert len(concatenation) == nb_others * 10
    loader = DataLoader(concatenation)
    for x, y, t in loader:
        pass


@pytest.mark.parametrize("nb_others", [0, 1, 2])
def test_concat_method(nb_others):
    x = np.random.rand(10, 2, 2, 3)
    y = np.ones((10,))
    t = np.ones((10,))

    base_set = TaskSet(x, y, t, None)
    initial_len = len(base_set)

    others = [
        TaskSet(np.copy(x), np.copy(y), np.copy(t), None) for _ in range(nb_others)
    ]
    base_set.concat(*others)
    assert len(base_set) == initial_len + nb_others * initial_len
    loader = DataLoader(base_set)
    for x, y, t in loader:
        pass


@pytest.mark.parametrize("val_split,nb_val", [(0., 0), (0.1, 1), (0.8, 8), (0.99, 9), (1.0, 10)])
def test_split_train_val(val_split, nb_val):
    x = np.random.rand(10, 2, 2, 3)
    y = np.ones((10,))
    t = np.ones((10,))

    base_set = TaskSet(x, y, t, None)

    train_set, val_set = split_train_val(base_set, val_split)
    assert len(val_set) == nb_val
    assert len(train_set) + len(val_set) == len(base_set)


def test_split_train_val_loading():
    x = np.random.rand(10, 2, 2, 3)
    y = np.ones((10,))
    t = np.ones((10,))

    base_set = TaskSet(x, y, t, None)

    train_set, val_set = split_train_val(base_set, 0.2)

    for task_set in (train_set, val_set):
        loader = DataLoader(task_set, batch_size=32)
        for x, y, t in loader:
            pass


@pytest.mark.parametrize("nb_samples", [1, 5, 10])
def test_get_random_samples(nb_samples):
    x = np.ones((10, 2, 2, 3))
    y = np.ones((10,))
    t = np.ones((10,))

    base_set = TaskSet(x, y, t, None)

    base_set.get_random_samples(nb_samples=nb_samples)


@pytest.mark.parametrize("nb_samples", [1, 5, 10])
def test_get_raw_samples(nb_samples):
    x = np.ones((10, 2, 2, 3))
    y = np.ones((10,))
    t = np.ones((10,))

    base_set = TaskSet(x, y, t, None)

    data, y_, t_ = base_set.get_raw_samples(indexes=range(nb_samples))

    assert (x[:nb_samples] == data).all()
    assert (y[:nb_samples] == y_).all()
    assert (t[:nb_samples] == t_).all()



import numpy as np
import pytest
from torch import nn
import torch
from copy import deepcopy
from continuum.metrics import Logger

from continuum.metrics import get_model_size


# yapf: disable

@pytest.fixture
def numpy_data():
    nb_classes = 20
    nb_tasks = 5
    nb_data = 100
    seen_classes = 0
    inc_classes = nb_classes // nb_tasks

    targets = []
    tasks = []

    for t in range(nb_tasks):
        seen_classes += inc_classes

        task_targets = np.concatenate([
            np.ones(nb_data) * c
            for c in range(seen_classes)
        ])
        task_tasks = np.concatenate([
            np.ones(nb_data * inc_classes) * tt
            for tt in range(t + 1)
        ])  # We also see previous data

        targets.append(task_targets)
        tasks.append(task_tasks)

    return targets, tasks


@pytest.fixture
def torch_models():
    class Small(nn.Module):
        def __init__(self):
            super().__init__()
            self.conv = nn.Sequential(nn.Conv2d(2, 3, 1, bias=False), nn.Conv2d(2, 3, 1, bias=False))
            self.fc = nn.Linear(5, 4)
            self.scalar = nn.Parameter(torch.tensor(3.))

    class Big(nn.Module):
        def __init__(self):
            super().__init__()
            self.conv = nn.Sequential(nn.Conv2d(2, 3, 1, bias=False), nn.Conv2d(2, 3, 1, bias=False))
            self.fc = nn.Linear(5, 4)
            self.fc2 = nn.Linear(5, 4)
            self.scalar = nn.Parameter(torch.tensor(3.))

    return Small(), Big()


def test_logger_nb_tasks(numpy_data):
    logger = Logger()
    all_targets, all_tasks = numpy_data
    nb_tasks = 3
    nb_epochs = 5
    for task in range(nb_tasks):
        for epoch in range(nb_epochs):
            for targets, task_ids in zip(all_targets, all_tasks):
                preds = np.copy(targets)
                logger.add([preds, targets, task_ids], subset="train")
            logger.end_epoch()
        logger.end_task()

    assert logger.nb_tasks == nb_tasks


def test_logger_simplest_add(numpy_data):
    logger = Logger()
    all_targets, all_tasks = numpy_data
    nb_tasks = 3
    nb_epochs = 5
    for task in range(nb_tasks):
        for epoch in range(nb_epochs):
            for targets, task_ids in zip(all_targets, all_tasks):
                preds = np.copy(targets)
                logger.add([preds, targets, task_ids], subset="train")
            logger.end_epoch()
        logger.end_task()


def test_logger_add_tensor(numpy_data):
    """
    test to check if we can use the logger to log random tensor with random keword
    """
    logger = Logger(list_keywords=['RandKeyword'])
    all_targets, all_tasks = numpy_data
    nb_tasks = 3
    nb_epochs = 5
    for task in range(nb_tasks):
        for epoch in range(nb_epochs):
            for targets, task_ids in zip(all_targets, all_tasks):
                rand_vector = torch.randn(15)
                logger.add(rand_vector, keyword='RandKeyword')
            logger.end_epoch()
        logger.end_task()


@pytest.mark.parametrize("mode,expected", [
    ("best", 1.), ("worst", 0.), ("random", None)
])
def test_metrics(numpy_data, mode, expected):
    logger = Logger()
    all_targets, all_tasks = numpy_data

    for targets, task_ids in zip(all_targets, all_tasks):
        if mode == "best":
            preds = np.copy(targets)
        elif mode == "worst":
            # Trick to never generate the correct predictions
            # only work for more three classes or more
            preds = (np.copy(targets) + 1) % np.max(targets)
        else:
            preds = np.random.randint(0, np.max(targets) + 1, targets.shape)

        logger.add(value=[preds, targets, task_ids], subset="train")
        logger.add(value=[preds, targets, task_ids], subset="test")

        accuracies = [
            logger.accuracy, logger.online_cumulative_performance,
            logger.accuracy_A,
        ]
        for acc in accuracies:
            if expected is not None:
                assert acc == expected, (acc, mode)
            assert 0. <= acc <= 1.
        assert -1. <= logger.backward_transfer <= 1.0
        assert -1. <= logger.forward_transfer <= 1.0
        assert 0. <= logger.positive_backward_transfer <= 1.0
        assert 0. <= logger.remembering <= 1.0
        assert 0. <= logger.forgetting <= 1.0

    assert 0. <= logger.average_incremental_accuracy <= 1.


@pytest.mark.parametrize("mode,expected", [
    ("best", 1.), ("worst", 0.), ("random", None)
])
def test_accuracy_per_task(numpy_data, mode, expected):
    logger = Logger()
    all_targets, all_tasks = numpy_data

    for targets, task_ids in zip(all_targets, all_tasks):
        if mode == "best":
            preds = np.copy(targets)
        elif mode == "worst":
            # Trick to never generate the correct predictions
            # only work for more three classes or more
            preds = (np.copy(targets) + 1) % np.max(targets)
        else:
            preds = np.random.randint(0, np.max(targets) + 1, targets.shape)
        logger.add(value=[preds, targets, task_ids], subset="test")

    accuracies = logger.accuracy_per_task

    for accuracy in accuracies:
        assert 0. <= accuracy <= 1.0


@pytest.mark.parametrize("batch_size", [
    1, 32, None
])
def test_online_accuracy(numpy_data, batch_size):
    logger = Logger()
    all_targets, _ = numpy_data
    targets = all_targets[0]

    # we check that when no data is in the logger online_accuracy generate an error
    check_raised(lambda: logger.online_accuracy)

    batch_size = batch_size or len(targets)
    for batch_index in range(0, len(targets), batch_size):
        y = targets[batch_index: batch_index + batch_size]
        x = np.copy(y)

        logger.add(value=[x, y, None], subset="train")
        logger.online_accuracy
    logger.online_accuracy

    logger.add(value=[targets, np.copy(targets), None], subset="train")
    logger.online_accuracy


def test_require_subset_test(numpy_data):
    logger = Logger()
    check_raised(lambda: logger.accuracy)

    values = [numpy_data[0][0], numpy_data[0][0], numpy_data[0][1]]
    logger.add(values, subset="test")
    logger.accuracy


def test_require_subset_train(numpy_data):
    logger = Logger()
    check_raised(lambda: logger.online_cumulative_performance)

    values = [numpy_data[0][0], numpy_data[0][0], numpy_data[0][1]]
    logger.add(values, subset="train")
    logger.online_cumulative_performance


def test_model_size(torch_models):
    small, big = torch_models
    assert get_model_size(small) < get_model_size(big)


def test_model_growth(torch_models):
    small, big = torch_models

    logger1 = Logger(list_keywords=['model_size'])  # Logger declaration with parameter name
    logger1.add(get_model_size(small), keyword='model_size')
    logger1.end_task()
    logger1.add(get_model_size(small), keyword='model_size')
    logger1.end_task()
    ms1 = logger1.model_size_growth

    logger2 = Logger(['model_size'])  # Logger declaration without parameter name
    logger2.add(get_model_size(small), keyword='model_size')
    logger2.end_task()
    logger2.add(get_model_size(big), keyword='model_size')
    logger2.end_task()
    ms2 = logger2.model_size_growth

    logger3 = Logger(['model_size'])
    logger3.add(get_model_size(big), keyword='model_size')
    logger3.end_task()
    logger3.add(get_model_size(small), keyword='model_size')
    logger3.end_task()
    ms3 = logger3.model_size_growth

    logger4 = Logger(['model_size'])
    logger4.add(get_model_size(big), keyword='model_size')
    logger4.end_task()
    logger4.add(get_model_size(big), keyword='model_size')
    logger4.end_task()
    ms4 = logger4.model_size_growth

    assert ms1 == ms4 == ms3 == 1.0
    assert 0. <= ms2 < 1.


@pytest.mark.slow
def test_example_doc():
    from torch.utils.data import DataLoader
    import numpy as np

    from continuum import ClassIncremental
    from continuum.datasets import MNIST
    from continuum.metrics import Logger

    train_scenario = ClassIncremental(
        MNIST(data_path="my/data/path", download=True, train=True),
        increment=2
    )
    test_scenario = ClassIncremental(
        MNIST(data_path="my/data/path", download=True, train=False),
        increment=2
    )

    # model = ...

    logger = Logger(list_subsets=['train', 'test'])

    for task_id, (train_taskset, test_taskset) in enumerate(zip(train_scenario, test_scenario)):
        train_loader = DataLoader(train_taskset)
        test_loader = DataLoader(test_taskset)

        for x, y, t in train_loader:
            predictions = y  # model(x)

            logger.add([predictions, y, None], subset="train")
            _ = (f"Online accuracy: {logger.online_accuracy}")

        for x_test, y_test, t_test in test_loader:
            preds_test = y_test

            logger.add([preds_test, y_test, t_test], subset="test")

        _ = (f"Task: {task_id}, acc: {logger.accuracy}, avg acc: {logger.average_incremental_accuracy}")
        _ = (f"BWT: {logger.backward_transfer}, FWT: {logger.forward_transfer}")

        logger.end_task()


def check_raised(func):
    has_raised = False
    try:
        func()
    except:
        has_raised = True
    finally:
        assert has_raised



import pytest
import torch
import numpy as np

from continuum.datasets import InMemoryDataset
from continuum import ClassIncremental
from continuum import rehearsal


@pytest.fixture
def scenario():
    x = np.random.randn(100, 2)
    y = np.concatenate([
        np.ones(10) * i
        for i in range(10)
    ])
    t = None

    dataset = InMemoryDataset(x, y, t)
    return ClassIncremental(dataset, increment=2)


@pytest.mark.parametrize("nb,method", [
    (10, rehearsal.herd_random),
    (20, rehearsal.herd_random),
    (5, rehearsal.herd_random),
    (10, rehearsal.herd_closest_to_barycenter),
    (20, rehearsal.herd_closest_to_barycenter),
    (5, rehearsal.herd_closest_to_barycenter),
    (10, rehearsal.herd_closest_to_cluster),
    (20, rehearsal.herd_closest_to_cluster),
    (5, rehearsal.herd_closest_to_cluster),
])
def test_nb_sampled(scenario, nb, method):
    for taskset in scenario:
        x, y, t = taskset.get_raw_samples()

        mx, my, mt = method(x, y, t, x, nb)
        assert len(mx) == 2 * min(nb, 10)



@pytest.mark.parametrize("memory_size,method,fixed", [
    (200, rehearsal.herd_random, True),
    (100, rehearsal.herd_random, True),
    (50, rehearsal.herd_random, True),
    (200, rehearsal.herd_random, False),
    (100, rehearsal.herd_random, False),
    (50, rehearsal.herd_random, False),
])
def test_memory(scenario, memory_size, method, fixed):
    memory = rehearsal.RehearsalMemory(
        memory_size, method, fixed, 10
    )
    assert len(memory) == 0

    c = 0
    for taskset in scenario:
        x, y, t = taskset.get_raw_samples()

        c += 2
        memory.add(x, y, t, x)
        assert memory.nb_classes == c
        if fixed:
            nb_per_class = min(memory_size, 100) // 10
            assert len(memory) == nb_per_class * c
        else:
            ideal_size = min(min(memory_size, 100), 10 * c)
            assert ideal_size >= len(memory) >= ideal_size - 5



@pytest.mark.parametrize("name,method", [
    ("random", rehearsal.herd_random),
    ("barycenter", rehearsal.herd_closest_to_barycenter),
    ("cluster", rehearsal.herd_closest_to_cluster),
])
def test_memory_name(name, method):
    memory = rehearsal.RehearsalMemory(
        20, name, True, 10
    )
    assert memory.herding_method == method




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as f:
    dependencies = list(map(lambda x: x.strip(), f.readlines()))


setuptools.setup(
    name="continuum",
    version="1.1.1",
    author="Arthur Douillard, Timothée Lesort",
    author_email="ar.douillard@gmail.com",
    description="A clean and simple library for Continual Learning in PyTorch.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Continvvm/continuum",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=dependencies
)



# from tqdm import tqdm
from torch.utils.data import Dataset
import numpy as np
import json
import os
from torchvision import transforms as tt
from PIL import Image
import torch
import multiprocessing
import sys
from continuum.datasets.base import InMemoryDataset
from functools import partial
import copy
import requests
import h5py


class Synbols(InMemoryDataset):
    categorical_attributes = ['char', 'font', 'alphabet', 'is_bold', 'is_slant']
    continuous_attributes = ['rotation', 'translation.x', 'translation.y', 'scale']
    def __init__(
            self,
            data_path: str,
            task_type: str = "char",
            domain_incremental_task: str = None,
            domain_increments: int = None,
            train: bool = True,
            dataset_name: str = "default_n=100000_2020-Oct-19.h5py",
            download: bool = True):
        """Wraps Synbols in a Continuum dataset for compatibility with Sequoia  

        Args:
            data_path (str): Path where the dataset will be saved
            train (bool): Whether to use the train split
            task_type (str): Options are 'char', and 'font'
            domain_incremental_task (str): The options are listed as static attribuets of the "Synbols" class.
            domain_increments (int): Amount of domain increments.
            dataset_name (str): See https://github.com/ElementAI/synbols-resources/raw/master/datasets/generated/',
            download (bool): Whether to download the dataset
        """
        assert(domain_incremental_task is None and domain_increments is None or \
            domain_incremental_task is not None and domain_increments is not None)

        if download:  # done here in order to pass x and y to super
            full_path = get_data_path_or_download("default_n=100000_2020-Oct-19.h5py",
                                                  data_root=data_path)
        else:
            full_path = os.path.join(data_path, dataset_name)

        data = SynbolsHDF5(full_path,
                           task_type,
                           domain_incremental_task=domain_incremental_task,
                           mask='random',
                           trim_size=None,
                           raw_labels=False)
        data = SynbolsSplit(data, 'train' if train else 'val', domain_incremental_task, domain_increments)

        super().__init__(data.x, data.y, data.task_id, train=train, download=download)


def _read_json_key(args):
    string, key = args
    return json.loads(string)[key]

def process_task(task, fields):
    data = json.loads(task)
    ret = []
    for field in fields:
        if '.x' in field:
            ret.append(data[field[:-2]][0])
        elif '.y' in field:
            ret.append(data[field[:-2]][1])
        else:
            ret.append(data[field])
    return ret

class SynbolsHDF5:
    """HDF5 Backend Class"""

    def __init__(self, path, task, domain_incremental_task=None, ratios=[0.6, 0.2, 0.2], mask=None, trim_size=None, raw_labels=False, reference_mask=None):
        """Constructor: loads data and parses labels.

        Args:
            path (str): path where the data is stored (see full_path above)
            task (str): 'char', 'font', or the field of choice 
            domain_incremental_task (str): task used for domain-incremental learning
            ratios (list, optional): The train/val/test split ratios. Defaults to [0.6, 0.2, 0.2].
            mask (ndarray, optional): Mask with the data partition. Defaults to None.
            trim_size (int, optional): Trim the dataset to a smaller size, for debugging speed. Defaults to None.
            raw_labels (bool, optional): Whether to include all the attributes of the synbols for each batch. Defaults to False.
            reference_mask (ndarray, optional): If train and validation are done with two different datasets, the 
                                                reference mask specifies the partition of the training data. Defaults to None.

        Raises:
            ValueError: Error message
        """
        self.path = path
        self.task = task
        self.domain_incremental_task = domain_incremental_task
        self.ratios = ratios
        print("Loading hdf5...")
        with h5py.File(path, 'r') as data:
            self.x = data['x'][...]
            y = data['y'][...]
            print("Converting json strings to labels...")
            parse_fields = [self.task]
            if self.domain_incremental_task is not None:
                parse_fields.append(self.domain_incremental_task)
            with multiprocessing.Pool(min(8, multiprocessing.cpu_count())) as pool:
                self.y = pool.map(partial(process_task, fields=parse_fields), y)
            self.y = list(map(np.array, zip(*self.y)))
            if self.domain_incremental_task is not None:
                self.y, self.domain_y = self.y
            else:
                self.y = self.y[0]
                self.domain_y = None
            print("Done converting.")

            self.mask = data["split"][mask][...]

            self.raw_labels = None

            self.trim_size = trim_size
            self.reference_mask = None
            print("Done reading hdf5.")

    def parse_mask(self, mask, ratios):
        return mask.astype(bool)

class SynbolsSplit(Dataset):
    def __init__(self, dataset, split, domain_incremental_task, domain_increments, transform=None):
        """Given a Backend (dataset), it splits the data in train, val, and test.


        Args:
            dataset (object): backend to load, it should contain the following attributes:
                - x, y, mask, ratios, path, task, mask
            split (str): train, val, or test
            domain_incremental_task (str): attribute used for domain-incremental learning
            domain_increments (int): number of domain increments
            transform (torchvision.transforms, optional): A composition of torchvision transforms. Defaults to None.
        """
        self.path = dataset.path
        self.task = dataset.task
        self.mask = dataset.mask
        self.domain_incremental_task = domain_incremental_task
        self.domain_increments = domain_increments
        if dataset.raw_labels is not None:
            self.raw_labelset = dataset.raw_labelset
        self.raw_labels = dataset.raw_labels
        self.ratios = dataset.ratios
        self.split = split
        if transform is None:
            self.transform = lambda x: x
        else:
            self.transform = transform
        self.split_data(dataset.x, dataset.y, dataset.mask, dataset.ratios, dataset.domain_y)

    def split_data(self, x, y, mask, ratios, domain_y, rng=np.random.RandomState(42)):
        if mask is None:
            if self.split == 'train':
                start = 0
                end = int(ratios[0] * len(x))
            elif self.split == 'val':
                start = int(ratios[0] * len(x))
                end = int((ratios[0] + ratios[1]) * len(x))
            elif self.split == 'test':
                start = int((ratios[0] + ratios[1]) * len(x))
                end = len(x)
            indices = rng.permutation(len(x))
            indices = indices[start:end]
        else:
            mask = mask[:, ["train", "val", "test"].index(self.split)]
            indices = np.arange(len(y))  # 0....nsamples
            indices = indices[mask]
        self.labelset = list(sorted(set(y)))
        self.y = np.array([self.labelset.index(y) for y in y])
        self.x = x[indices]
        self.y = self.y[indices]
        if self.raw_labels is not None:
            self.raw_labels = np.array(self.raw_labels)[indices]

        # Create "domain_increments" bins from the attribute used for domain incremental learning 
        if self.domain_increments is not None:
            if self.domain_incremental_task in Synbols.categorical_attributes:
                self.domains = list(sorted(set(domain_y)))
                self.tasks_per_domain = len(self.domains) // self.domain_increments
                self.domain_labels = list(range(self.domain_increments)) * self.tasks_per_domain + list(range(self.domain_increments))[:(len(self.domains) % self.domain_increments)]
                self.task_id = domain_y[indices]
                self.task_id = np.array([self.domain_labels[self.domains.index(d_y)] for d_y in self.task_id])
            elif self.domain_incremental_task in Synbols.continuous_attributes:
                self.domains = np.linspace(domain_y.min(), domain_y.max() + 1e-4, self.domain_increments + 1)
                domain_y = domain_y[indices]
                self.task_id = np.zeros(len(domain_y), dtype=int)
                for i in range(1, self.domain_increments):
                    self.task_id[(domain_y >= self.domains[i - 1]) & (domain_y < self.domains[i])] = i
            else:
                raise ValueError("Domain attribute not found")
            
            if len(set(self.task_id)) != self.domain_increments:
                raise RuntimeError("""The number of tasks differs from the number of domain increments. 
                                        This could happen if the number of domains in the dataset is less than 
                                        the requested number of increments. For instance, requesting 2000 
                                        font increments for a dataset with 1500 fonts will result in failure""")
        else:
            self.task_id = None

    def __getitem__(self, item):
        if self.raw_labels is None:
            return self.transform(self.x[item]), self.y[item]
        else:
            return self.transform(self.x[item]), self.y[item], self.raw_labels[item]

    def __len__(self):
        return len(self.x)


def get_data_path_or_download(dataset, data_root):
    """Finds a dataset locally and downloads if needed.

    Args:
        dataset (str): dataset name. For instance 'camouflage_n=100000_2020-Oct-19.h5py'.
            See https://github.com/ElementAI/synbols-resources/tree/master/datasets/generated for the complete list. (please ignore .a[a-z] extensions)
        data_root (str): path where the dataset will be or is stored. If empty string, it defaults to $TMPDIR 

    Raises:
        ValueError: dataset name does not exist in local path nor in remote

    Returns:
        str: dataset final path 
    """
    url = 'https://zenodo.org/record/4701316/files/%s?download=1' % dataset
    if data_root == "":
        data_root = os.environ.get("TMPDIR", "/tmp")
    full_path = os.path.join(data_root, dataset)

    if os.path.isfile(full_path):
        print("%s found." % full_path)
        return full_path
    else:
        print("Downloading %s..." % full_path)

    r = requests.head(url)
    is_big = not r.ok

    if is_big:
        r = requests.head(url + ".aa")
        if not r.ok:
            raise ValueError("Dataset %s" % dataset, "not found in remote.")
        response = input("Download more than 3GB (Y/N)?: ").lower()
        while response not in ["y", "n"]:
            response = input("Download more than 3GB (Y/N)?: ").lower()
        if response == "n":
            print("Aborted")
            sys.exit(0)
        parts = []
        current_part = "a"
        while r.ok:
            r = requests.head(url + ".a%s" % current_part)
            parts.append(url + ".a" + current_part)
            current_part = chr(ord(current_part) + 1)
    else:
        parts = [url]

    if not os.path.isfile(full_path):
        with open(full_path, 'wb') as file:
            for i, part in enumerate(parts):
                print("Downloading part %d/%d" % (i + 1, len(parts)))

                # Streaming, so we can iterate over the response.
                response = requests.get(part, stream=True)
                total_size_in_bytes = int(
                    response.headers.get('content-length', 0))
                block_size = 1024  # 1 Kilobyte
                # progress_bar = tqdm(total=total_size_in_bytes,
                                    # unit='iB', unit_scale=True)
                for data in response.iter_content(block_size):
                    # progress_bar.update(len(data))
                    file.write(data)
                # progress_bar.close()
                # if total_size_in_bytes != 0:# and progress_bar.n != total_size_in_bytes:
                    # print("ERROR, something went wrong downloading %s" % url)
    return full_path



# pylint: disable=C0401
# flake8: noqa
from continuum.datasets.base import (
    ImageFolderDataset, InMemoryDataset, PyTorchDataset, _ContinuumDataset
)
from continuum.datasets.cifar100 import CIFAR100
from continuum.datasets.core50 import (Core50, Core50v2_79, Core50v2_196, Core50v2_391)
from continuum.datasets.fellowship import (CIFARFellowship, Fellowship, MNISTFellowship)
from continuum.datasets.imagenet import ImageNet100, ImageNet1000, TinyImageNet200
from continuum.datasets.synbols import Synbols
from continuum.datasets.nlp import MultiNLI
from continuum.datasets.pytorch import (
    CIFAR10, EMNIST, KMNIST, MNIST, QMNIST, FashionMNIST
)
from continuum.datasets.svhn import SVHN
from continuum.datasets.cub200 import CUB200
from continuum.datasets.awa2 import AwA2
from continuum.datasets.pascalvoc import PascalVOC2012
from continuum.datasets.stream51 import Stream51
from continuum.datasets.dtd import DTD
from continuum.datasets.rainbow_mnist import RainbowMNIST
from continuum.datasets.ctrl import CTRL, CTRLplus, CTRLminus, CTRLin, CTRLout, CTRLplastic



import warnings
from copy import copy
from typing import Callable, List, Union

import numpy as np

from continuum.datasets import _ContinuumDataset
from continuum.scenarios import _BaseScenario


class ContinualScenario(_BaseScenario):
    """Continual Loader, generating datasets for the consecutive tasks.

    Scenario: the scenario is entirely defined by the task label vector in the cl_dataset

    :param cl_dataset: A continual dataset.
    :param transformations: A list of transformations applied to all tasks. If
                            it's a list of list, then the transformation will be
                            different per task.
    """

    def __init__(
            self,
            cl_dataset: _ContinuumDataset,
            transformations: Union[List[Callable], List[List[Callable]]] = None,
    ) -> None:
        self.check_data(cl_dataset)
        super().__init__(cl_dataset=cl_dataset, nb_tasks=self.nb_tasks, transformations=transformations)

    def check_data(self, cl_dataset: _ContinuumDataset):
        x, y, t = cl_dataset.get_data()

        assert t is not None, print("The t vector should be defined for this scenario")
        assert len(x) == len(y) == len(t), print("data, label and task label vectors need to have the same length")

        list_unique_tasks_ids = np.unique(t)
        self._nb_tasks = len(list_unique_tasks_ids)
        self.dataset = (x, y, t)

        # we order the list to check if all index from 0 to nb_tasks are in the list
        list_unique_tasks_ids.sort()

        assert np.all(list_unique_tasks_ids == np.arange(self.nb_tasks)), print(
            f"there should be at least one data point"
            f"for each task from task id equal"
            f"zero to num_tasks-1 \n (unique task indexes) {list_unique_tasks_ids} vs "
            f"(expected) {np.arange(self.nb_tasks)}")


    #nothing to do in the setup function
    def _setup(self, nb_tasks: int) -> int:
        return nb_tasks



import warnings
from typing import Callable, List, Union

import numpy as np
import torch
from torchvision import transforms

from continuum.datasets import _ContinuumDataset
from continuum.scenarios import TransformationIncremental


class Permutations(TransformationIncremental):
    """Continual Loader, generating datasets for the consecutive tasks.

    Scenario: Permutations scenarios, use same data for all task but with pixels permuted.
    Each task get a specific permutation, such as all tasks are different.

    :param cl_dataset: A continual dataset.
    :param nb_tasks: The scenario's number of tasks.
    :param base_transformations: List of transformations to apply to all tasks.
    :param seed: initialization seed for the permutations.
    :param shared_label_space: If true same data with different transformation have same label
    """

    def __init__(
        self,
        cl_dataset: _ContinuumDataset,
        nb_tasks: Union[int, None] = None,
        base_transformations: List[Callable] = None,
        seed: Union[int, List[int]] = 0,
        shared_label_space=True
    ):
        trsfs = self._generate_transformations(seed, nb_tasks)

        super().__init__(
            cl_dataset=cl_dataset,
            incremental_transformations=trsfs,
            base_transformations=base_transformations,
            shared_label_space=shared_label_space
        )

    def _generate_transformations(self, seed, nb_tasks):
        if isinstance(seed, int):
            if nb_tasks is None:
                raise ValueError("You must specify a number of tasks if a single seed is provided.")
            rng = np.random.RandomState(seed=seed)
            seed = rng.permutation(100000)[:nb_tasks - 1]
        elif nb_tasks is not None and nb_tasks != len(seed) + 1:
            warnings.warn(
                f"Because a list of seed was provided {seed}, "
                f"the number of tasks is automatically set to "
                f"len(number of seeds) + 1 = {len(seed) + 1}"
            )

        return [PermutationTransform(seed=None)] + [PermutationTransform(seed=int(s)) for s in seed]

    def get_task_transformation(self, task_index):
        return transforms.Compose(self.trsf.transforms + [self.inc_trsf[task_index]])


class PermutationTransform:
    """Permutation transformers.

    This transformer is initialized with a seed such as same seed = same permutation.
    Seed 0 means no permutations

    :param seed: seed to initialize the random number generator
    """

    def __init__(self, seed: Union[int, None]):
        self.seed = seed
        self.g_cpu = torch.Generator()

    def __call__(self, x):
        shape = list(x.shape)
        x = x.reshape(-1)
        # if seed is None, no permutations
        if self.seed is not None:
            self.g_cpu.manual_seed(self.seed)
            perm = torch.randperm(x.numel(), generator=self.g_cpu).long()
            x = x[perm]
        return x.reshape(shape)



from typing import Tuple

import numpy as np
from torchvision import datasets as torchdata
from torchvision import transforms

from continuum.datasets import PyTorchDataset


cifar100_coarse_labels = np.array([4, 1, 14, 8, 0, 6, 7, 7, 18, 3,
                                   3, 14, 9, 18, 7, 11, 3, 9, 7, 11,
                                   6, 11, 5, 10, 7, 6, 13, 15, 3, 15,
                                   0, 11, 1, 10, 12, 14, 16, 9, 11, 5,
                                   5, 19, 8, 8, 15, 13, 14, 17, 18, 10,
                                   16, 4, 17, 4, 2, 0, 17, 4, 18, 17,
                                   10, 3, 2, 12, 12, 16, 12, 1, 9, 19,
                                   2, 10, 0, 1, 16, 12, 9, 13, 15, 13,
                                   16, 19, 2, 4, 6, 19, 5, 5, 8, 19,
                                   18, 1, 2, 15, 6, 0, 17, 8, 14, 13])


class CIFAR100(PyTorchDataset):
    """Continuum use of the CIFAR100 dataset.
    We adapt this dataset to create scenario from the combination of both of category labels and class labels.

    References:
        * "Learning Multiple Layers of Features from Tiny Images"
          Alex Krizhevsky

    :param data_path: The folder path containing the data.
    :param download: An option useless in this case.
    :param labels_type: labels type define if we use class labels or category labels for each data point.
    :param task_labels: labels type define what type of labels we use if we want to create a task id vector.
    """

    def __init__(self, *args, labels_type: str = "class", task_labels: str = None, **kwargs):
        super().__init__(*args, dataset_type=torchdata.cifar.CIFAR100, **kwargs)
        if not labels_type in ["class", "category"]:
            AssertionError("unknown labels_type parameter, choose among ['class', 'category']")
        if not  task_labels in [None, "class", "category"]:
            AssertionError("unknown task_labels parameter, choose among [None, 'class', 'category']")

        self.labels_type = labels_type
        self.task_labels_type = task_labels

        if self.task_labels_type is None:
            # the dataset does not provide a task id vector
            self.t = None
        elif self.task_labels_type == "class":
            # the dataset provides a task id vector based on classes
            self.t = self.dataset.targets
        elif self.task_labels_type == "category":
            # the dataset provides a task id vector based on categories
            self.t = cifar100_coarse_labels[self.dataset.targets]

        if self.labels_type == "category":
            # here we replace class labels by category labels to annotate data.

            # Classes labels from 0-19 (instead of 0 to 99)
            # cf : "superclasses" or "coarse labels" https://www.cs.toronto.edu/~kriz/cifar.html

            # code from https://github.com/ryanchankh/cifar100coarse/blob/master/cifar100coarse.py
            self.dataset.targets = cifar100_coarse_labels[self.dataset.targets]

            # update classes
            self.dataset.classes = [['beaver', 'dolphin', 'otter', 'seal', 'whale'],
                                    ['aquarium_fish', 'flatfish', 'ray', 'shark', 'trout'],
                                    ['orchid', 'poppy', 'rose', 'sunflower', 'tulip'],
                                    ['bottle', 'bowl', 'can', 'cup', 'plate'],
                                    ['apple', 'mushroom', 'orange', 'pear', 'sweet_pepper'],
                                    ['clock', 'keyboard', 'lamp', 'telephone', 'television'],
                                    ['bed', 'chair', 'couch', 'table', 'wardrobe'],
                                    ['bee', 'beetle', 'butterfly', 'caterpillar', 'cockroach'],
                                    ['bear', 'leopard', 'lion', 'tiger', 'wolf'],
                                    ['bridge', 'castle', 'house', 'road', 'skyscraper'],
                                    ['cloud', 'forest', 'mountain', 'plain', 'sea'],
                                    ['camel', 'cattle', 'chimpanzee', 'elephant', 'kangaroo'],
                                    ['fox', 'porcupine', 'possum', 'raccoon', 'skunk'],
                                    ['crab', 'lobster', 'snail', 'spider', 'worm'],
                                    ['baby', 'boy', 'girl', 'man', 'woman'],
                                    ['crocodile', 'dinosaur', 'lizard', 'snake', 'turtle'],
                                    ['hamster', 'mouse', 'rabbit', 'shrew', 'squirrel'],
                                    ['maple_tree', 'oak_tree', 'palm_tree', 'pine_tree', 'willow_tree'],
                                    ['bicycle', 'bus', 'motorcycle', 'pickup_truck', 'train'],
                                    ['lawn_mower', 'rocket', 'streetcar', 'tank', 'tractor']]

    def get_data(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        x, y, _ = super().get_data()
        return x, y, self.t

    @property
    def transformations(self):
        """Default transformations if nothing is provided to the scenario."""
        return [transforms.ToTensor(),
                transforms.Normalize((0.5071, 0.4867, 0.4408), (0.2675, 0.2565, 0.2761))]



import os
from typing import Tuple, Union

from torchvision import datasets as torchdata
import numpy as np

from continuum.datasets import ImageFolderDataset
from continuum.download import download, unzip


class ImageNet1000(ImageFolderDataset):
    """ImageNet1000 dataset.

    Simple wrapper around ImageFolderDataset to provide a link to the download
    page.
    """
    @property
    def transformations(self):
        """Default transformations if nothing is provided to the scenario."""
        return [transforms.ToTensor(),
                transforms.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225))]


    def _download(self):
        if not os.path.exists(self.data_path):
            raise IOError(
                "You must download yourself the ImageNet dataset."
                " Please go to http://www.image-net.org/challenges/LSVRC/2012/downloads and"
                " download 'Training images (Task 1 & 2)' and 'Validation images (all tasks)'."
            )
        print("ImageNet already downloaded.")


class ImageNet100(ImageNet1000):
    """Subset of ImageNet1000 made of only 100 classes.

    You must download the ImageNet1000 dataset then provide the images subset.
    If in doubt, use the option at initialization `download=True` and it will
    auto-download for you the subset ids used in:
        * Small Task Incremental Learning
          Douillard et al. 2020
    """

    train_subset_url = "https://github.com/Continvvm/continuum/releases/download/v0.1/train_100.txt"
    test_subset_url = "https://github.com/Continvvm/continuum/releases/download/v0.1/val_100.txt"

    def __init__(
        self, *args, data_subset: Union[Tuple[np.array, np.array], str, None] = None, **kwargs
    ):
        self.data_subset = data_subset
        super().__init__(*args, **kwargs)

    def _download(self):
        super()._download()

        filename = "val_100.txt"
        self.subset_url = self.test_subset_url
        if self.train:
            filename = "train_100.txt"
            self.subset_url = self.train_subset_url

        if self.data_subset is None:
            self.data_subset = os.path.join(self.data_path, filename)
            download(self.subset_url, self.data_path)

    def get_data(self) -> Tuple[np.ndarray, np.ndarray, Union[np.ndarray, None]]:
        data = self._parse_subset(self.data_subset, train=self.train)  # type: ignore
        return (*data, None)

    def _parse_subset(
        self,
        subset: Union[Tuple[np.array, np.array], str, None],
        train: bool = True
    ) -> Tuple[np.array, np.array]:
        if isinstance(subset, str):
            x, y = [], []

            with open(subset, "r") as f:
                for line in f:
                    split_line = line.split(" ")
                    path = split_line[0].strip()
                    x.append(os.path.join(self.data_path, path))
                    y.append(int(split_line[1].strip()))
            x = np.array(x)
            y = np.array(y)
            return x, y
        return subset  # type: ignore


class TinyImageNet200(ImageFolderDataset):
    """Smaller version of ImageNet.

    - 200 classes
    - 500 images per class
    - size 64x64
    """

    url = "http://cs231n.stanford.edu/tiny-imagenet-200.zip"

    def _download(self):
        path = os.path.join(self.data_path, "tiny-imagenet-200")
        if not os.path.exists(f"{path}.zip"):
            download(self.url, self.data_path)
        if not os.path.exists(path):
            unzip(f"{path}.zip")

        print("TinyImagenet is downloaded.")

    def get_data(self) -> Tuple[np.ndarray, np.ndarray, Union[np.ndarray, None]]:
        return self._format(
            torchdata.ImageFolder(
                os.path.join(self.data_path, "tiny-imagenet-200", "train" if self.train else "val")
            ).imgs
        )



from typing import List, Tuple, Union

from PIL import Image
from skimage.transform import resize
import numpy as np

from continuum.datasets import CIFAR10, SVHN, FashionMNIST, MNIST, DTD, RainbowMNIST
from continuum.datasets import _ContinuumDataset


class CTRL(_ContinuumDataset):
    """CTRL generic class.

    Reference:
        * Efficient Continual Learning with Modular Networks and Task-Driven Priors
          Tom Veniat, Ludovic Denoyer, Marc'Aurelio Ranzato
          ICLR 2021

    To get a feeling about this class, please look at the actual implementation of
    the various instances of CTRL datasets (e.g. CTRLminus, CTRLplus, etc.)

    :param datasets: The list of continual datasets to use.
    :param target_size: The common image size for all.
    :param split: Which split among train/val/test.
    :param proportions: How much images (in absolute value or %) to take per dataset.
    :param class_counter: The initial class counter per dataset, helps when seeing
                          twice the same dataset we want them to have the same or
                          different labels.
    :param class_subsets: A subset of classes to sample from the i-th dataset.
    :param seed: A random seed for reproducibility.
    """
    def __init__(
        self,
        datasets: List[_ContinuumDataset],
        target_size: Tuple[int, int] = (32, 32),
        split: str = "train",
        proportions: Union[None, List[int], List[float]] = None,
        class_counter: Union[None, List[int]] = None,
        class_subsets: Union[None, List[int]] = None,
        seed: int = 1
    ):
        class_counter = class_counter or [0 for _ in range(len(datasets))]
        class_subsets = class_subsets or [None for _ in range(len(datasets))]
        proportions = proportions or [None for _ in range(len(datasets))]
        if len(proportions) != len(datasets):
            raise ValueError(
                f"There must have as much proportions {len(proportions)} as datasets ({len(datasets)})."
            )
        if len(class_counter) != len(datasets):
            raise ValueError(
                f"There must have as much class counters {len(class_counter)} as datasets ({len(datasets)})."
            )
        if split not in ("train", "val", "test"):
            raise ValueError(f"Split must be train, val, or test; not {split}.")

        self.datasets = datasets
        self.target_size = target_size
        self.split = split
        self.proportions = proportions
        self.class_counter = class_counter
        self.class_subsets = class_subsets
        self.seed = seed

        super().__init__()

    def get_data(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        all_x, all_y, all_t = [], [], []

        task_id = 0
        for i, dataset in enumerate(self.datasets):
            x, y, _ = dataset.get_data()

            if self.class_subsets[i] is not None:
                indexes = np.where(np.isin(y, self.class_subsets[i]))[0]
                x, y = x[indexes], y[indexes]
            if self.split in ("train", "val") and self.proportions[i]:
                indexes = self.balanced_sampling(y, self.proportions[i], self.seed, self.split)
                x, y = x[indexes], y[indexes]
            if x.dtype == "S255":  # String paths
                x = self.open_and_resize(x, self.target_size)
            if len(x.shape) == 3:  # Grayscale images
                x = x[..., None]
            if x.shape[-1] == 1:  # Grayscale images
                x = np.repeat(x, 3, axis=-1)
            if x.shape[1:3] != self.target_size:  # Not the correct widh or height
                x = self.resize(x, self.target_size)

            all_x.append(x)
            all_y.append(y + self.class_counter[i])
            all_t.append(np.ones(len(y)) * task_id)
            task_id += 1

        return np.concatenate(all_x), np.concatenate(all_y), np.concatenate(all_t)

    def open_and_resize(self, paths: np.ndarray, size: Tuple[int, int]):
        x = np.zeros((len(paths), *size, 3), dtype=np.uint8)
        for i, path in enumerate(paths):
            img = Image.open(path).convert('RGB').resize(size)
            x[i] = np.array(img).astype(np.uint8)
        return x

    def resize(self, arrays: np.ndarray, size: Tuple[int, int]):
        new_arrays = np.zeros((len(arrays), *size, 3), dtype=np.uint8)
        for i, arr in enumerate(arrays):
            new_arrays[i] = resize(arr, size, preserve_range=True).astype(np.uint8)
        return new_arrays

    def balanced_sampling(self, y: np.ndarray, amount: Union[float, int], seed: int, split: str = "train"):
        """Samples a certain amount of data equally per class."""
        if isinstance(amount, float):
            amount = int(len(y) * amount)
        unique_classes = np.unique(y)
        if len(unique_classes) > amount:
            raise ValueError(
                f"Not enough amount ({amount}) for the number of classes ({len(unique_classes)})."
            )
        # There can be a few images lost, but that's not very important.
        amount_per_class = int(amount / len(unique_classes))

        rng = np.random.RandomState(seed=seed)

        indexes = []
        for c in unique_classes:
            class_indexes = np.where(y == c)[0]
            rng.shuffle(class_indexes)

            if split == "train":
                indexes.append(class_indexes[:amount_per_class])
            else:  # val
                indexes.append(class_indexes[-amount_per_class:])

        return np.concatenate(indexes)



class CTRLminus(CTRL):  # S^-
    def __init__(self, data_path: str = "", split: str = "train", download: bool = True, seed: int = 1):
        if split not in ("train", "val", "test"):
            raise ValueError(f"Split must be train, val, or test; not {split}.")
        train = split in ("train", "val")

        datasets = [
            CIFAR10(data_path=data_path, train=train, download=download),
            MNIST(data_path=data_path, train=train, download=download),
            DTD(data_path=data_path, train=train, download=download),
            FashionMNIST(data_path=data_path, train=train, download=download),
            SVHN(data_path=data_path, train=train, download=download),
            CIFAR10(data_path=data_path, train=train, download=download)
        ]

        if split == "train":
            proportions = [4000, 400, 400, 400, 400, 400]
        elif split == "val":
            proportions = [2000, 200, 200, 200, 200, 200]
        else:
            proportions = None

        super().__init__(
            datasets=datasets,
            proportions=proportions,
            class_counter=[0, 10, 20, 67, 77, 0],
            seed=seed,
            split=split
        )


class CTRLplus(CTRL):  # S^+
    def __init__(self, data_path: str = "", split: str = "train", download: bool = True, seed: int = 1):
        if split not in ("train", "val", "test"):
            raise ValueError(f"Split must be train, val, or test; not {split}.")
        train = split in ("train", "val")

        datasets = [
            CIFAR10(data_path=data_path, train=train, download=download),
            MNIST(data_path=data_path, train=train, download=download),
            DTD(data_path=data_path, train=train, download=download),
            FashionMNIST(data_path=data_path, train=train, download=download),
            SVHN(data_path=data_path, train=train, download=download),
            CIFAR10(data_path=data_path, train=train, download=download)
        ]

        if split == "train":
            proportions = [400, 400, 400, 400, 400, 4000]
        elif split == "val":
            proportions = [200, 200, 200, 200, 200, 2000]
        else:
            proportions = None

        super().__init__(
            datasets=datasets,
            proportions=proportions,
            class_counter=[0, 10, 20, 67, 77, 0],
            seed=seed,
            split=split
        )


class CTRLin(CTRL):  # S^{in}
    def __init__(self, data_path: str = "", split: str = "train", download: bool = True, seed: int = 1):
        if split not in ("train", "val", "test"):
            raise ValueError(f"Split must be train, val, or test; not {split}.")
        train = split in ("train", "val")

        color1, color2 = np.random.RandomState(seed=seed).choice(["red", "blue", "green"], 2)
        datasets = [
            RainbowMNIST(data_path=data_path, train=train, download=download, color=color1),
            CIFAR10(data_path=data_path, train=train, download=download),
            DTD(data_path=data_path, train=train, download=download),
            FashionMNIST(data_path=data_path, train=train, download=download),
            SVHN(data_path=data_path, train=train, download=download),
            RainbowMNIST(data_path=data_path, train=train, download=download, color=color2)
        ]

        if split == "train":
            proportions = [4000, 400, 400, 400, 400, 50]
        elif split == "val":
            proportions = [2000, 200, 200, 200, 200, 30]
        else:
            proportions = None

        super().__init__(
            datasets=datasets,
            proportions=proportions,
            class_counter=[0, 10, 20, 67, 77, 0],
            seed=seed,
            split=split
        )


class CTRLout(CTRL):  # S^{out}
    def __init__(self, data_path: str = "", split: str = "train", download: bool = True, seed: int = 1):
        if split not in ("train", "val", "test"):
            raise ValueError(f"Split must be train, val, or test; not {split}.")
        train = split in ("train", "val")

        datasets = [
            CIFAR10(data_path=data_path, train=train, download=download),
            MNIST(data_path=data_path, train=train, download=download),
            DTD(data_path=data_path, train=train, download=download),
            FashionMNIST(data_path=data_path, train=train, download=download),
            SVHN(data_path=data_path, train=train, download=download),
            CIFAR10(data_path=data_path, train=train, download=download)
        ]

        if split == "train":
            proportions = [4000, 400, 400, 400, 400, 400]
        elif split == "val":
            proportions = [2000, 200, 200, 200, 200, 200]
        else:
            proportions = None

        super().__init__(
            datasets=datasets,
            proportions=proportions,
            class_counter=[0, 10, 20, 67, 77, 87],
            seed=seed,
            split=split
        )


class CTRLplastic(CTRL):  # S^{pl}
    def __init__(self, data_path: str = "", split: str = "train", download: bool = True, seed: int = 1):
        if split not in ("train", "val", "test"):
            raise ValueError(f"Split must be train, val, or test; not {split}.")
        train = split in ("train", "val")

        datasets = [
            MNIST(data_path=data_path, train=train, download=download),
            DTD(data_path=data_path, train=train, download=download),
            FashionMNIST(data_path=data_path, train=train, download=download),
            SVHN(data_path=data_path, train=train, download=download),
            CIFAR10(data_path=data_path, train=train, download=download)
        ]

        if split == "train":
            proportions = [400, 400, 400, 400, 4000]
        elif split == "val":
            proportions = [200, 200, 200, 200, 2000]
        else:
            proportions = None

        super().__init__(
            datasets=datasets,
            proportions=proportions,
            class_counter=[0, 10, 57, 67, 77],
            seed=seed,
            split=split
        )


# TODO: add CTRLlong



import os
from typing import Tuple

import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split
from torchvision import datasets as torchdata

from continuum.datasets import ImageFolderDataset
from continuum.download import download, untar


class AwA2(ImageFolderDataset):
    """AwA2 dataset.

    * Zero-Shot Learning - A Comprehensive Evaluation of the Good, the Bad and the Ugly
      Y. Xian, C. H. Lampert, B. Schiele, Z. Akata
      TPAMI 2018
    """
    images_url = "https://cvml.ist.ac.at/AwA2/AwA2-data.zip"
    split_v2_url = "http://datasets.d2.mpi-inf.mpg.de/xian/xlsa17.zip"

    def __init__(self, data_path, train: bool = True, download: bool = True, test_split: float = 0.2):
        self._attributes = None
        self.test_split = test_split
        super().__init__(data_path, train, download)

    @property
    def attributes(self):
        if self._attributes is None:
            att = np.loadtxt(
                os.path.join(
                    self.data_path, "Animals_with_Attributes2",
                    "predicate-matrix-continuous.txt"
                )
            )
            self._attributes = att / np.linalg.norm(att, axis=-1, keepdims=True)

        return self._attributes

    @property
    def data_type(self):
        return "image_data_path"

    def _download(self):
        if not os.path.exists(os.path.join(self.data_path, "Animals_with_Attributes2")):
            zip_path = os.path.join(self.data_path, "AwA2-data.zip")

            if not os.path.exists(zip_path):
                print("Downloading zip images archive...", end=' ')
                download(self.images_url, zip_path)
                print('Done!')

            print('Extracting archive...', end=' ')
            untar(zip_path)
            print('Done!')

        if not os.path.exists(os.path.join(self.data_path, "xlsa17")):
            zip_path = os.path.join(self.data_path, "xlsa17.zip")

            if not os.path.exists(zip_path):
                print("Downloading zip split archive...", end=' ')
                download(self.split_v2_url, zip_path)
                print('Done!')

            print('Extracting archive...', end=' ')
            untar(zip_path)
            print('Done!')

    def get_data(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        dataset = torchdata.ImageFolder(os.path.join(self.data_path, "Animals_with_Attributes2", "JPEGImages"))
        x, y, _ = self._format(dataset.imgs)

        x_train, x_test, y_test, y_train = train_test_split(
            x, y,
            test_size=self.test_split,
            random_state=1
        )

        if self.train:
            return x_train, y_train, None
        return x_test, y_test, None



import os
import copy

import numpy as np
import pytest

from continuum.scenarios import Rotations
from continuum.datasets import MNIST, CIFAR100, InMemoryDataset
from tests.test_classorder import InMemoryDatasetTest


DATA_PATH = os.environ.get("CONTINUUM_DATA_PATH")

@pytest.fixture
def numpy_data():
    nb_classes = 6
    nb_data = 100

    x_train = []
    y_train = []
    for i in range(nb_classes):
        x_train.append(np.ones((nb_data, 4, 4, 3), dtype=np.uint8) * i)
        y_train.append(np.ones(nb_data) * i)
    x_train = np.concatenate(x_train)
    y_train = np.concatenate(y_train)

    x_test = np.copy(x_train)
    y_test = np.copy(y_train)

    return (x_train, y_train.astype(int)), (x_test, y_test.astype(int))


'''
Test the initialization with three tasks
'''


def test_init(numpy_data):
    train, test = numpy_data
    dummy = InMemoryDatasetTest(*train)

    Trsf_0 = 0
    Trsf_1 = (15, 20)
    Trsf_2 = 45

    list_degrees = [Trsf_0, Trsf_1, Trsf_2]

    scenario = Rotations(cl_dataset=dummy, nb_tasks=3, list_degrees=list_degrees)

    for task_id, train_dataset in enumerate(scenario):
        continue


@pytest.mark.parametrize("shared_label_space", [True, False])
def test_shared_labels(numpy_data, shared_label_space):
    train, test = numpy_data
    dummy = InMemoryDatasetTest(*train)
    list_degrees = [0, 15, 45]
    nb_classes = 6

    scenario = Rotations(cl_dataset=dummy, nb_tasks=3, list_degrees=list_degrees, shared_label_space=shared_label_space)

    for task_id, taskset in enumerate(scenario):
        classes = taskset.get_classes()

        if shared_label_space:
            assert (classes == np.arange(nb_classes)).all()
        else:
            assert (classes == np.arange(nb_classes) + task_id * nb_classes).all(), task_id


def test_fail_init(numpy_data):
    train, test = numpy_data
    dummy = InMemoryDatasetTest(*train)

    Trsf_0 = 2
    Trsf_1 = (15, 20, 25)  # non sense
    Trsf_2 = 45

    list_degrees = [Trsf_0, Trsf_1, Trsf_2]

    # should detect that a transformation is non-sens in the list
    with pytest.raises(ValueError):
        Rotations(cl_dataset=dummy, nb_tasks=3, list_degrees=list_degrees)


@pytest.mark.slow
@pytest.mark.parametrize("shared_label_space", [True, False])
@pytest.mark.parametrize("dataset", [MNIST, CIFAR100])
def test_with_dataset(dataset, shared_label_space):
    dataset = dataset(data_path=DATA_PATH, download=True, train=True)
    list_degrees = [0, 45, 90]
    scenario = Rotations(cl_dataset=dataset,
                          nb_tasks=3,
                          list_degrees=list_degrees,
                          shared_label_space=shared_label_space)
    assert len(scenario) == 3

    previous_classes = None
    for task_id, taskset in enumerate(scenario):
        classes = taskset.get_classes()

        if task_id > 0:
            if shared_label_space:
                assert (classes == previous_classes).all()
            else:
                assert (classes == previous_classes + len(classes)).all(), classes

        previous_classes = classes



import numpy as np
import pytest
import torch
from PIL import Image
from torchvision.transforms import transforms

from continuum.datasets import InMemoryDataset
from continuum.scenarios import TransformationIncremental

NB_CLASSES = 6


@pytest.fixture
def numpy_data():
    nb_data = 100  # not too small to have all classes

    x_train = []
    y_train = []
    x_train.append(
        np.array([np.random.randint(100, size=(2, 2, 3)).astype(dtype=np.uint8)] * nb_data)
    )
    y_train.append(np.random.randint(NB_CLASSES, size=(nb_data)))
    x_train = np.concatenate(x_train)
    y_train = np.concatenate(y_train)

    return x_train, y_train.astype(int)


'''
Test the initialization with three tasks
'''


def test_init(numpy_data):
    x, y = numpy_data
    dummy = InMemoryDataset(x, y, train='train')

    Trsf_0 = []
    Trsf_1 = [transforms.RandomAffine(degrees=[45, 45])]
    Trsf_2 = [transforms.RandomAffine(degrees=[90, 90])]

    list_transf = [Trsf_0, Trsf_1, Trsf_2]

    scenario = TransformationIncremental(
        cl_dataset=dummy, incremental_transformations=list_transf
    )

    ref_data = None
    raw_ref_data = None
    for task_id, taskset in enumerate(scenario):

        samples, _, _ = taskset.get_random_samples(10)
        # we need raw data to apply same transformation as the TransformationIncremental class
        raw_samples, _, _ = taskset.get_raw_samples(range(10))

        if task_id == 0:
            ref_data = samples
            raw_ref_data = raw_samples
        else:
            # we verify that data has changed
            assert not torch.all(ref_data.eq(samples))

            assert (raw_samples == raw_ref_data
                    ).all()  # raw data should be the same in this scenario

            # we test transformation on one data point and verify if it is applied
            trsf = list_transf[task_id][0]
            raw_sample = Image.fromarray(raw_ref_data[0].astype("uint8"))
            trsf_data = trsf(raw_sample)
            trsf_data = transforms.ToTensor()(trsf_data)

            assert torch.all(trsf_data.eq(samples[0]))


'''
Test the initialization with three tasks with degree range
'''


def test_init_range(numpy_data):
    x, y = numpy_data
    dummy = InMemoryDataset(x, y)

    Trsf_0 = []
    Trsf_1 = [transforms.RandomAffine(degrees=[40, 50])]
    Trsf_2 = [transforms.RandomAffine(degrees=[85, 95])]

    list_transf = [Trsf_0, Trsf_1, Trsf_2]

    scenario = TransformationIncremental(
        cl_dataset=dummy, incremental_transformations=list_transf
    )


@pytest.mark.parametrize("shared_label_space", [False, True])
def test_init_shared_label_space(numpy_data, shared_label_space):
    x, y = numpy_data
    dummy = InMemoryDataset(x, y)

    Trsf_0 = []
    Trsf_1 = [transforms.RandomAffine(degrees=[40, 50])]
    Trsf_2 = [transforms.RandomAffine(degrees=[85, 95])]

    dummy_transf = [Trsf_0, Trsf_1, Trsf_2]

    scenario = TransformationIncremental(
        cl_dataset=dummy,
        incremental_transformations=dummy_transf,
        shared_label_space=shared_label_space
    )

    for task_id, taskset in enumerate(scenario):
        assert taskset.nb_classes == NB_CLASSES
        classes = taskset.get_classes()
        if shared_label_space:
            assert classes.max() == NB_CLASSES - 1
            assert classes.min() == 0
        else:
            assert classes.max() == (NB_CLASSES * (task_id + 1)) - 1
            assert classes.min() == (NB_CLASSES * task_id)


def test_get_task_transformation(numpy_data):
    x, y = numpy_data
    dummy = InMemoryDataset(x, y)

    Trsf_0 = []
    Trsf_1 = [transforms.RandomAffine(degrees=[40, 50])]
    Trsf_2 = [transforms.RandomAffine(degrees=[85, 95])]

    dummy_transf = [Trsf_0, Trsf_1, Trsf_2]

    base_transformations = [
        transforms.ToTensor(),
        transforms.Normalize((0.1307,), (0.3081,))
    ]

    scenario = TransformationIncremental(
        cl_dataset=dummy,
        incremental_transformations=dummy_transf,
        base_transformations=base_transformations
    )

    for task_id, taskset in enumerate(scenario):
        # first task specific transformation then base_transformation
        tot_transf_task = transforms.Compose(dummy_transf[task_id] + base_transformations)

        # we compare the str representation of the composition
        assert tot_transf_task.__repr__() == scenario.get_task_transformation(task_id).__repr__()


def test_init_fail2(numpy_data):
    train = numpy_data
    dummy = InMemoryDataset(*train)

    # No transformation is set
    with pytest.raises(TypeError):
        scenario = TransformationIncremental(cl_dataset=dummy)


def test_indexing():
    x = np.zeros((20, 2, 2, 3), dtype=np.uint8)
    x[:, 0, 0] = 1  # add a 1 on the top-left

    y = np.ones((20,), dtype=np.int32)

    dataset = InMemoryDataset(x, y)

    trsfs = [
        [_discrete_rotation(0)],
        [_discrete_rotation(1)],
        [_discrete_rotation(2)],
        [_discrete_rotation(3)],
    ]
    scenario = TransformationIncremental(
        cl_dataset=dataset,
        incremental_transformations=trsfs
    )

    for task_id in range(len(scenario)):
        task_set = scenario[task_id]

        x, _, t = task_set[0]
        _check_rotation(x, task_id)


@pytest.mark.parametrize("indexes_slice", [
    slice(0, 1), slice(0, 3),
    slice(0, 4, 2)
])
def test_advanced_indexing(indexes_slice):
    """
    This code creates dummy images of 2x2 all likewise:
    [
        1 0
        0 0
    ]

    Then we apply discrete rotations to produce the four possible variations
    (1 on the top-right, bottom-right, bottom-left in addition of the original
    top-left). We then sample multiple tasks together and check that the associated
    task label of the sample matches the rotations it was applied to.
    """
    x = np.zeros((20, 2, 2, 3), dtype=np.uint8)
    x[:, 0, 0] = 1  # add a 1 on the top-left

    y = np.ones((20,), dtype=np.int32)

    dataset = InMemoryDataset(x, y)

    trsfs = [
        [_discrete_rotation(0)],
        [_discrete_rotation(1)],
        [_discrete_rotation(2)],
        [_discrete_rotation(3)],
    ]
    scenario = TransformationIncremental(
        cl_dataset=dataset,
        incremental_transformations=trsfs
    )
    start = indexes_slice.start if indexes_slice.start is not None else 0
    stop = indexes_slice.stop if indexes_slice.stop is not None else len(scenario) + 1
    step = indexes_slice.step if indexes_slice.step is not None else 1
    task_index = set(list(range(start, stop, step)))

    task_set = scenario[indexes_slice]
    seen_tasks = set()

    for i in range(len(task_set)):
        x, _, t = task_set[i]
        _check_rotation(x, t)
        seen_tasks.add(t)
    assert seen_tasks == task_index


def _discrete_rotation(rot):
    def _fun(x):
        if rot == 0:
            one = (0, 0)
        elif rot == 1:
            one = (0, 1)
        elif rot == 2:
            one = (1, 1)
        elif rot == 3:
            one = (1, 0)

        x = np.array(x)
        x.fill(0)
        x[one[0], one[1], :] = 1
        return Image.fromarray(x.astype(np.uint8))
    return _fun


def _check_rotation(x, rot):
    if rot == 0:
        one = (0, 0)
    elif rot == 1:
        one = (0, 1)
    elif rot == 2:
        one = (1, 1)
    elif rot == 3:
        one = (1, 0)
    else:
        assert False, rot

    for i in range(2):
        for j in range(2):
            if (i, j) == one:
                v = 1
            else:
                v = 0
            assert int(255 * x[0, i, j]) == v, (x[0, i, j], rot, (i, j), v)
